"""
Variable constants common to this project
"""

__all__ = ["COLUMNS_TO_DE_IDENTIFY",
           "MODIFIED_OMOP_PORTION_DIR_MAC",
           "MODIFIED_OMOP_PORTION_DIR_WIN",
           "NOTES_PORTION_DIR_MAC",
           "NOTES_PORTION_DIR_WIN",
           "OLD_MAPS_DIR_PATH",
           "OMOP_PORTION_DIR_MAC",
           "OMOP_PORTION_DIR_WIN"]

import re
from pathlib import Path
# Local packages
from drapi.code.drapi.constants.constants import DATA_TYPES_DICT
from drapi.code.drapi.constants.phiVariables import (LIST_OF_PHI_VARIABLES_BO,
                                          LIST_OF_PHI_VARIABLES_I2B2,
                                          LIST_OF_PHI_VARIABLES_NOTES,
                                          LIST_OF_PHI_VARIABLES_OMOP,
                                          VARIABLE_NAME_TO_FILE_NAME_DICT,
                                          VARIABLE_SUFFIXES_BO,
                                          VARIABLE_SUFFIXES_I2B2,
                                          VARIABLE_SUFFIXES_NOTES,
                                          VARIABLE_SUFFIXES_OMOP)
from drapi.code.drapi.drapi import flatExtend

# Project parameters: Meta-variables
STUDY_TYPE = "Limited Data Set (LDS)"  # Pick from "Non-Human", "Limited Data Set (LDS)", "Identified"
IRB_NUMBER = "IRB202300703"  # TODO
DATA_REQUEST_ROOT_DIRECTORY_DEPTH = 3  # TODO  # NOTE To prevent unexpected results, like moving, writing, or deleting the wrong files, set this to folder that is the immediate parent of concatenated result and the intermediate results folder.


# Project parameters: Aliases: Definitions  # TODO
# NOTE: Some variable names are not standardized. This section is used by the de-identification process when looking for the de-identification map. This way several variables can be de-identified with the same map. If you have variables with a custom, non-BO name, you should alias them, if necessary using the following format:
# VAR_ALIASES_CUSTOM_VARS = {"Custom Variable 1": "BO Equivalent 1",
#                            "Custom Variable 2": "BO Equivalent 2"}
VAR_ALIASES_NOTES_ENCOUNTERS = {"EncounterCSN": "Encounter # (CSN)",
                                "EncounterKey": "Encounter Key"}
VAR_ALIASES_NOTES_NOTES = {"NoteID": "Note ID",
                           "NoteKey": "Note Key",
                           "LinkageNoteID": "Linkage Note ID"}
VAR_ALIASES_NOTES_PATIENTS = {"MRN_GNV": "MRN (UF)",
                              "MRN_JAX": "MRN (Jax)",
                              "PatientKey": "Patient Key"}
VAR_ALIASES_NOTES_PROVIDERS = {"AuthoringProviderKey": "Provider Key",
                               "AuthorizingProviderKey": "Provider Key",
                               "CosignProviderKey": "Provider Key",
                               "OrderingProviderKey": "Provider Key",
                               "ProviderKey": "Provider Key"}
VAR_ALIASES_SDOH_ENCOUNTERS = {"NOTE_ENCNTR_KEY": "Linkage Note ID",
                               "NOTE_KEY": "Note Key"}
LIST_OF_ALIAS_DICTS = [VAR_ALIASES_NOTES_ENCOUNTERS,
                       VAR_ALIASES_NOTES_NOTES,
                       VAR_ALIASES_NOTES_PATIENTS,
                       VAR_ALIASES_NOTES_PROVIDERS,
                       VAR_ALIASES_SDOH_ENCOUNTERS]
VARIABLE_ALIASES = {}
for di in LIST_OF_ALIAS_DICTS:
    VARIABLE_ALIASES.update(di)

# Project parameters: Aliases: Data types  # TODO
# NOTE Add each variable in `VARIABLE_ALIASES` to `ALIAS_DATA_TYPES`.
ALIAS_DATA_TYPES_MANUAL = {}  # TODO
ALIAS_DATA_TYPES_AUTOMATIC = {}
for alias, variableName in VARIABLE_ALIASES.items():
    if alias not in DATA_TYPES_DICT.keys():
        ALIAS_DATA_TYPES_AUTOMATIC[alias] = DATA_TYPES_DICT[variableName]
    else:
        pass
ALIAS_DATA_TYPES = ALIAS_DATA_TYPES_MANUAL.copy()
ALIAS_DATA_TYPES.update(ALIAS_DATA_TYPES_AUTOMATIC)
DATA_TYPES_DICT.update(ALIAS_DATA_TYPES)

# Project parameters: Aliases: Alias-to-File-name conversion  # TODO
VARIABLES_TO_OVER_WRITE_MANUALLY = {}  # TODO
VARIABLES_TO_ADD_FROM_ALIASES = {variableName: variableName for variableName in VARIABLE_ALIASES.keys()}
VARIABLES_TO_OVER_WRITE_LIST = [VARIABLES_TO_ADD_FROM_ALIASES,
                                VARIABLES_TO_OVER_WRITE_MANUALLY]  # NOTE The manual list must go last, to over-write the automatic additions.
VARIABLES_TO_ADD_DICT = {}
for di in VARIABLES_TO_OVER_WRITE_LIST:
    VARIABLES_TO_ADD_DICT.update(di)
VARIABLE_NAME_TO_FILE_NAME_DICT.update(VARIABLES_TO_ADD_DICT)
FILE_NAME_TO_VARIABLE_NAME_DICT = {fileName: varName for varName, fileName in VARIABLE_NAME_TO_FILE_NAME_DICT.items()}

# QA: Make sure all aliases have data types.
li = []
for alias in VARIABLE_ALIASES.keys():
    if alias in DATA_TYPES_DICT.keys():
        pass
    else:
        li.append(alias)
if len(li) > 0:
    string = "\n".join(sorted(li))
    msg = f"""The following variable aliases have no data type assigned:\n{string}"""
    raise Exception(msg)

# Add aliases to list of variables to de-identify by adding the keys from `VARIABLE_ALIASES` and `ALIAS_DATA_TYPES` to `COLUMNS_TO_DE_IDENTIFY`.
LIST_OF_PHI_VARIABLES_FROM_ALIASES = [variableName for variableName in VARIABLE_ALIASES.keys()] + [variableName for variableName in ALIAS_DATA_TYPES.keys()]
LIST_OF_PHI_VARIABLES_FROM_ALIASES = list(set(LIST_OF_PHI_VARIABLES_FROM_ALIASES))

# Project parameters: Columns to de-identify
if STUDY_TYPE.lower() == "Non-Human":
    LIST_OF_PHI_VARIABLES_TO_KEEP = []
else:
    LIST_OF_PHI_VARIABLES_TO_KEEP = []
COLUMNS_TO_DE_IDENTIFY = flatExtend([LIST_OF_PHI_VARIABLES_BO,
                                     LIST_OF_PHI_VARIABLES_I2B2,
                                     LIST_OF_PHI_VARIABLES_NOTES,
                                     LIST_OF_PHI_VARIABLES_OMOP,
                                     LIST_OF_PHI_VARIABLES_FROM_ALIASES])
COLUMNS_TO_DE_IDENTIFY = [variableName for variableName in COLUMNS_TO_DE_IDENTIFY if variableName not in LIST_OF_PHI_VARIABLES_TO_KEEP]

# Project parameters: Variable suffixes
VARIABLE_SUFFIXES_LIST = [VARIABLE_SUFFIXES_BO,
                          VARIABLE_SUFFIXES_I2B2,
                          VARIABLE_SUFFIXES_NOTES,
                          VARIABLE_SUFFIXES_OMOP]
VARIABLE_SUFFIXES = dict()
for variableSuffixDict in VARIABLE_SUFFIXES_LIST:
    VARIABLE_SUFFIXES.update(variableSuffixDict)

# Project parameters: Portion directories
MODIFIED_OMOP_PORTION_DIR_MAC = Path("data/output/convertColumns/...")  # TODO
MODIFIED_OMOP_PORTION_DIR_WIN = Path(r"data\output\convertColumns\...")  # TODO

NOTES_ROOT_DIRECTORY = Path(r"..\Notes\original_note_pull\data\Output")  # TODO
NOTES_PORTION_DIR_MAC = NOTES_ROOT_DIRECTORY.joinpath("free_text")
NOTES_PORTION_DIR_WIN = NOTES_ROOT_DIRECTORY.joinpath("free_text")

OMOP_PORTION_DIR_MAC = Path(r"..\OMOP_Structured_Data\data\output\identified")  # TODO
OMOP_PORTION_DIR_WIN = Path(r"..\OMOP_Structured_Data\data\output\identified")  # TODO

OMOP_PORTION_2_DIR_MAC = Path(r"..\Intermediate Results\OMOP Portion\data\output\identified\2024-01-23 13-54-26")  # TODO
OMOP_PORTION_2_DIR_WIN = Path(r"..\Intermediate Results\OMOP Portion\data\output\identified\2024-01-23 13-54-26")  # TODO

# Project parameters: File criteria
NOTES_PORTION_FILE_CRITERIA = [lambda pathObj: pathObj.suffix.lower() == ".csv",
                               lambda pathObj: True if re.search(pattern=r"metadata_\d+.csv",
                                                         string=pathObj.name) else False]
OMOP_PORTION_FILE_CRITERIA = [lambda pathObj: pathObj.suffix.lower() == ".csv"]


# Project parameters: Maps
OLD_MAPS_DIR_PATH = {"EncounterCSN": [NOTES_ROOT_DIRECTORY.joinpath("mapping/map_encounter.csv")],
                     "LinkageNoteID": [NOTES_ROOT_DIRECTORY.joinpath("mapping/map_note_link.csv")],
                     "NoteKey": [NOTES_ROOT_DIRECTORY.joinpath("mapping/map_note.csv")],
                     "OrderKey": [NOTES_ROOT_DIRECTORY.joinpath("mapping/map_order.csv")],
                     "PatientKey": [NOTES_ROOT_DIRECTORY.joinpath("mapping/map_patient.csv")],
                     "ProviderKey": [NOTES_ROOT_DIRECTORY.joinpath("mapping/map_provider.csv")]}

# Quality assurance
if __name__ == "__main__":
    ALL_VARS = [MODIFIED_OMOP_PORTION_DIR_MAC,
                MODIFIED_OMOP_PORTION_DIR_WIN,
                NOTES_ROOT_DIRECTORY,
                NOTES_PORTION_DIR_MAC,
                NOTES_PORTION_DIR_WIN,
                OMOP_PORTION_DIR_MAC,
                OMOP_PORTION_DIR_WIN]

    for li in OLD_MAPS_DIR_PATH.values():
        ALL_VARS.extend(li)

    for path in ALL_VARS:
        print(path.exists())



"""
De-identify files

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
# NOTE Expects all files to be CSV files. This is because it uses "pd.read_csv".
# TODO Assign portion name to each path (per OS) so that portion files are stored in their respective folders, this prevents file from being overwritten in the unlikely, but possible, case files from different portions have the same name.
# TODO Investigate if a symlink can be made for files that are copied without alteration, to save space and time on larger projects.
"""

# Third-party packages
import pandas as pd
# Local packages
from drapi.drapi import getTimestamp, makeDirPath


def deleteColumns(COLUMNS_TO_DELETE,
                  COLUMNS_TO_DELETE_DICT,
                  LIST_OF_PORTION_CONDITIONS,
                  CHUNK_SIZE,
                  listOfPortionDirs,
                  pipelineOutputDir,
                  logger,
                  rootDirectory,
                  ROOT_DIRECTORY):
    """
    TODO Columns to delete. The list variable deletes column matches in any file. The dictionary variable delete the columns that are in the file named in the dictionary key. Below are typical values for a limited data release.
    """
    functionName = __name__.split(".")[-1]
    runOutputDir = pipelineOutputDir.joinpath(functionName, getTimestamp())
    makeDirPath(runOutputDir)
    logger.info(f"""Begin running "{functionName}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Function arguments:

    # Arguments
    ``: "{""}"
    """)

    # De-identify columns
    logger.info("""Deleting columns not authorized for release.""")
    for directory, fileConditions in zip(listOfPortionDirs, LIST_OF_PORTION_CONDITIONS):
        # Act on directory
        logger.info(f"""Working on directory "{directory.absolute().relative_to(rootDirectory)}".""")
        for file in directory.iterdir():
            logger.info(f"""  Working on file "{file.absolute().relative_to(rootDirectory)}".""")
            conditions = [condition(file) for condition in fileConditions]
            if all(conditions):
                # Set file options
                exportPath = runOutputDir.joinpath(file.name)
                fileMode = "w"
                fileHeaders = True
                # Read file
                logger.info("""    File has met all conditions for processing.""")
                numChunks = sum([1 for _ in pd.read_csv(file, chunksize=CHUNK_SIZE)])
                dfChunks = pd.read_csv(file, chunksize=CHUNK_SIZE)
                for it, dfChunk in enumerate(dfChunks, start=1):
                    dfChunk = pd.DataFrame(dfChunk)
                    # Work on chunk
                    logger.info(f"""  ..  Working on chunk {it} of {numChunks}.""")
                    for columnName in dfChunk.columns:
                        # Work on column
                        logger.info(f"""  ..    Working on column "{columnName}".""")
                        if file.stem in COLUMNS_TO_DELETE_DICT.keys():
                            listOfColumns = COLUMNS_TO_DELETE + COLUMNS_TO_DELETE_DICT[file.stem]
                        else:
                            listOfColumns = COLUMNS_TO_DELETE
                        if columnName in listOfColumns:
                            logger.info("""  ..  ..  Column must be deleted. Deleting column.""")
                            dfChunk = dfChunk.drop(columns=columnName)
                    # Save chunk
                    dfChunk.to_csv(exportPath, mode=fileMode, header=fileHeaders, index=False)
                    fileMode = "a"
                    fileHeaders = False
                    logger.info(f"""  ..  Chunk saved to "{exportPath.absolute().relative_to(rootDirectory)}".""")
            else:
                logger.info("""    This file does not need to be processed.""")

    # Clean up
    # TODO If input directory is empty, delete
    # TODO Delete intermediate run directory

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{functionName}".""")



"""
Convert OMOP person IDs to IDR patient keys.

# NOTE Does not expect data in nested directories (e.g., subfolders of "free_text"). Therefore it uses "Path.iterdir" instead of "Path.glob('*/**')".
"""

import logging
import sys
from pathlib import Path
# Third-party packages
import pandas as pd
from pandas.errors import EmptyDataError
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    personIDs2patientKeys,
                                    readDataFile,
                                    successiveParents)
from common import (DATA_REQUEST_ROOT_DIRECTORY_DEPTH,
                    OMOP_PORTION_DIR_MAC,
                    OMOP_PORTION_DIR_WIN,
                    OMOP_PORTION_FILE_CRITERIA)

# Arguments
COLUMNS_TO_CONVERT = ["person_id"]

MAC_PATHS = [OMOP_PORTION_DIR_MAC]
WIN_PATHS = [OMOP_PORTION_DIR_WIN]

LIST_OF_PORTION_CONDITIONS = [OMOP_PORTION_FILE_CRITERIA]

SETS_PATH = None

CHUNK_SIZE = 50000

# Arguments: Meta-variables
CONCATENATED_RESULTS_DIRECTORY_DEPTH = DATA_REQUEST_ROOT_DIRECTORY_DEPTH - 1
PROJECT_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH  # The concatenation suite of scripts is considered to be the "project".
IRB_DIR_DEPTH = CONCATENATED_RESULTS_DIRECTORY_DEPTH + 2
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "DATA_REQUEST_DIRECTORY"  # TODO One of the following:
                                           # ["IDR_DATA_REQUEST_DIRECTORY",      # noqa
                                           #  "IRB_DIRECTORY",                   # noqa
                                           #  "DATA_REQUEST_DIRECTORY",          # noqa
                                           #  "CONCATENATED_RESULTS_DIRECTORY"]  # noqa

LOG_LEVEL = "INFO"

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_ROOT_DIRECTORY_DEPTH)
IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    intermediateDataDir = dataDir.joinpath("intermediate")
    outputDataDir = dataDir.joinpath("output")
    if intermediateDataDir:
        runIntermediateDataDir = intermediateDataDir.joinpath(thisFileStem, runTimestamp)
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "CONCATENATED_RESULTS_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir

# Variables: Path construction: OS-specific
isAccessible = all([path.exists() for path in MAC_PATHS]) or all([path.exists() for path in WIN_PATHS])
if isAccessible:
    # If you have access to either of the below directories, use this block.
    operatingSystem = sys.platform
    if operatingSystem == "darwin":
        listOfPortionDirs = MAC_PATHS[:]
    elif operatingSystem == "win32":
        listOfPortionDirs = WIN_PATHS[:]
    else:
        raise Exception("Unsupported operating system")
else:
    # If the above option doesn't work, manually copy the database to the `input` directory.
    print("Not implemented. Check settings in your script.")
    sys.exit()

# Directory creation: General
makeDirPath(runIntermediateDataDir)
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

if __name__ == "__main__":
    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(LOG_LEVEL)
    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)

    logging.basicConfig(format="[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s",
                        handlers=[fileHandler, streamHandler],
                        level=LOG_LEVEL)

    logging.info(f"""Begin running "{thisFilePath}".""")
    logging.info(f"""All other paths will be reported in debugging relative to `IRBDir`: "{IRBDir}".""")

    # Get set of values
    if SETS_PATH:
        logging.info(f"""Using the set of values previously collected from "{SETS_PATH}".""")
    else:
        logging.info("""Getting the set of values for each variable to convert.""")
        columnSetsVarsDi = {columnName: {"fpath": runIntermediateDataDir.joinpath(f"{columnName}.txt"),
                                         "fileMode": "w"} for columnName in COLUMNS_TO_CONVERT}
        for directory, fileConditions in zip(listOfPortionDirs, LIST_OF_PORTION_CONDITIONS):
            # Act on directory
            logging.info(f"""Working on directory "{directory.relative_to(IRBDir)}".""")
            for file in directory.iterdir():
                logging.info(f"""  Working on file "{file.absolute().relative_to(IRBDir)}".""")
                conditions = [condition(file) for condition in fileConditions]
                if all(conditions):
                    # Read file
                    logging.info("""    File has met all conditions for processing.""")
                    numChunks = sum([1 for _ in readDataFile(file, chunksize=CHUNK_SIZE)])
                    dfChunks = readDataFile(file, chunksize=CHUNK_SIZE)
                    for it, dfChunk in enumerate(dfChunks, start=1):
                        logging.info(f"""  ..  Working on chunk {it} of {numChunks}.""")
                        for columnName in dfChunk.columns:
                            logging.info(f"""  ..    Working on column "{columnName}".""")
                            if columnName in COLUMNS_TO_CONVERT:
                                logging.info("""  ..  ..  Column must be converted. Collecting values.""")
                                valuesSet = sorted(list(set(dfChunk[columnName].dropna().values)))
                                columnSetFpath = columnSetsVarsDi[columnName]["fpath"]
                                columnSetFileMode = columnSetsVarsDi[columnName]["fileMode"]
                                with open(columnSetFpath, columnSetFileMode) as file:
                                    for value in valuesSet:
                                        file.write(str(value))
                                        file.write("\n")
                                columnSetsVarsDi[columnName]["fileMode"] = "a"
                                logging.info(f"""  ..  ..  Values saved to "{columnSetFpath.absolute().relative_to(IRBDir)}" in the project directory.""")
                else:
                    logging.info("""    This file does not need to be processed.""")

    # Map values
    if SETS_PATH:
        setsPathDir = SETS_PATH
    else:
        setsPathDir = runIntermediateDataDir
    for file in setsPathDir.iterdir():
        columnName = file.stem
        logging.info(f"""  Working on variable "{columnName}" located at "{file.absolute().relative_to(IRBDir)}".""")
        # Read file
        try:
            df = pd.read_table(file, header=None)
        except EmptyDataError as err:
            _ = err
            df = pd.DataFrame()
        # Assert
        if df.shape[1] == 1:
            # Try to convert to integer-type
            try:
                df.iloc[:, 0] = df.iloc[:, 0].astype(int)
            except ValueError as err:
                _ = err
            # Check length differences
            len0 = len(df)
            values = set(df.iloc[:, 0].values)
            len1 = len(values)
            logging.info(f"""    The length of the ID array was reduced from {len0:,} to {len1:,} when removing duplicates.""")
        elif df.shape[1] == 0:
            pass
        # Map contents
        map_ = personIDs2patientKeys(list(values))
        # Save map
        mapPath = runOutputDir.joinpath(f"{columnName} map.csv")
        map_.to_csv(mapPath, index=False)
        logging.info(f"""    PersonID-to-PatientKey map saved to "{mapPath.absolute().relative_to(IRBDir)}".""")

    # Clean up
    # TODO If input directory is empty, delete
    # TODO Delete intermediate run directory

    # Output location summary
    logging.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(IRBDir)}".""")

    # End script
    logging.info(f"""Finished running "{thisFilePath.absolute().relative_to(IRBDir)}".""")



from __future__ import annotations

import concurrent.futures
import logging
import os
import random
import re
import shutil
import sys
from datetime import timedelta
from logging import Logger
from pathlib import Path
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from _typeshed.dbapi import DBAPIConnection
# Third-party packages
import pandas as pd
import pymssql
import sqlalchemy as sa
# Local packages
from drapi.code.drapi.drapi import (getTimestamp,
                                    makeDirPath,
                                    replace_sql_query,
                                    successiveParents)

# Arguments: Script arguments
COHORT_NAME = ""                                    # An arbitrary name used in file names
COHORT_FILE = ""                                    # A file name that is located directory specified by the variable `data_dir`
IRB_NUMBER = ""                                     # Used for creating the de-identification map IDs.
ID_TYPE = ""                                        # Pick from "EncounterCSN", "EncounterKey", or "PatientKey". Choose the ID type you used in `COHORT_FILE`
NOTE_VERSION = ""                                   # Pick from "all", or "last"
DE_IDENTIFICATION_MODE = ""                         # Pick from "deid", "lds", or "phi"
LOG_LEVEL = ""                                      # See the "logging" module for valid values for the `loglevel` parameter.
SQL_ENCOUNTER_EFFECTIVE_DATE_START = ''   # The beginning of date range of encounters to collect. Format: YYYY-MM-DD
SQL_ENCOUNTER_EFFECTIVE_DATE_END = ''     # The end of date range of encounters to collect. Format: YYYY-MM-DD

# Arguments: SQL connection settings
USE_WINDOWS_AUTHENTICATION = None                   # Boolean or `None`. By default this is determined based on the operating system, but can be forced by the user.
SERVER = "DWSRSRCH01.shands.ufl.edu"
DATABASE_PROD = "DWS_PROD"
DATABASE_NOTES = "DWS_NOTES"
USERDOMAIN = "UFAD"
USERNAME = os.environ["USER"]
UID = None
PWD = os.environ["HFA_UFADPWD"]

# Arguments: Meta-variables
PROJECT_DIR_DEPTH = 2
DATA_REQUEST_DIR_DEPTH = PROJECT_DIR_DEPTH + 2
IRB_DIR_DEPTH = DATA_REQUEST_DIR_DEPTH + 0
IDR_DATA_REQUEST_DIR_DEPTH = IRB_DIR_DEPTH + 3

ROOT_DIRECTORY = "PROJECT_OR_PORTION_DIRECTORY"  # TODO One of the following:
                                                 # ["IDR_DATA_REQUEST_DIRECTORY",    # noqa
                                                 #  "IRB_DIRECTORY",                 # noqa
                                                 #  "DATA_REQUEST_DIRECTORY",        # noqa
                                                 #  "PROJECT_OR_PORTION_DIRECTORY"]  # noqa

# Variables: Path construction: General
runTimestamp = getTimestamp()
thisFilePath = Path(__file__)
thisFileStem = thisFilePath.stem
projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
IRBDir, _ = successiveParents(thisFilePath, IRB_DIR_DEPTH)
IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
dataDir = projectDir.joinpath("data")
if dataDir:
    inputDataDir = dataDir.joinpath("input")
    outputDataDir = dataDir.joinpath("output")
    if outputDataDir:
        runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
logsDir = projectDir.joinpath("logs")
if logsDir:
    runLogsDir = logsDir.joinpath(thisFileStem)
sqlDir = projectDir.joinpath("sql")

if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
    rootDirectory = projectDir
elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
    rootDirectory = dataRequestDir
elif ROOT_DIRECTORY == "IRB_DIRECTORY":
    rootDirectory = IRBDir
elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
    rootDirectory = IDRDataRequestDir
else:
    raise Exception("An unexpected error occurred.")

# Variables: Path Construction: Project-specific
notes_dir = runOutputDir.joinpath('free_text')  # all notes related files are saved in 'notes' subdirectory of 'data' directory
map_dir = runOutputDir.joinpath('mapping')  # mappings are saved in 'mapping' subdirectory of 'data' folder.
disclosure_dir = runOutputDir.joinpath('disclosure')

# Variables: Map legacy variables to DRAPI-LEMUR standard variables: Script settings
cohort = COHORT_NAME
id_type = ID_TYPE
note_version = NOTE_VERSION
irb = IRB_NUMBER
deid_mode = DE_IDENTIFICATION_MODE

# Variables: Map legacy variables to DRAPI-LEMUR standard variables: Path construction
base_dir = str(projectDir)
data_dir = dataDir
sql_dir = sqlDir

# Variables: Map legacy variables to DRAPI-LEMUR standard variables: SQL Parameters
host = SERVER
database_prod = DATABASE_PROD
database_notes = DATABASE_NOTES

# Variables: SQL connection settings
if UID:
    uid = UID[:]
else:
    uid = fr"{USERDOMAIN}\{USERNAME}"
if isinstance(USE_WINDOWS_AUTHENTICATION, type(None)):
    operatingSystem = sys.platform
    if operatingSystem in ["win32"]:
        useWindowsAuthentication = True
    elif operatingSystem in ["darwin", "linux"]:
        useWindowsAuthentication = False
    else:
        raise Exception(f"""Unsupported operating system: "{operatingSystem}".""")
else:
    useWindowsAuthentication = USE_WINDOWS_AUTHENTICATION

# Directory creation: General
makeDirPath(runOutputDir)
makeDirPath(runLogsDir)

# Directory creation: Project-specific
for dir in [data_dir, sql_dir, notes_dir, disclosure_dir]:
    makeDirPath(dir)
if DE_IDENTIFICATION_MODE.lower() == "phi":
    pass
elif DE_IDENTIFICATION_MODE.lower() in ["deid", "lds"]:
    makeDirPath(map_dir)
else:
    raise Exception(f"Unexpected value for `DE_IDENTIFICATION_MODE`: {DE_IDENTIFICATION_MODE}.")

# Functions

def choosePathToLog(path: Path, rootPath: Path) -> Path:
    """
    Decides if a path is a subpath of `rootPath`. If it is, display it reltaive to `rootPath`. If it is not, display it as an absolute path.
    """
    commonPath = os.path.commonpath([path.absolute(), rootPath.absolute()])

    lenCommonPath = len(commonPath)
    lenRootPath = len(str(rootPath.absolute()))
    if lenCommonPath < lenRootPath:
        pathToDisplay = path
    elif lenCommonPath >= lenRootPath:
        pathToDisplay = path.absolute().relative_to(rootPath)
    else:
        raise Exception("An unexpected error occurred.")
                                    
    return pathToDisplay


def connectToDatabase(host: str,
                      database: str,
                      useWindowsAuthentication=True) -> DBAPIConnection:
    """
    Connects to a SQL database given a `host` (server) and `database` value.
    """
    if useWindowsAuthentication:
        connection = pymssql.connect(host=host,
                                     database=database)
    else:
        conStr = f"mssql+pymssql://{uid}:{PWD}@{host}/{database}"
        connection = sa.create_engine(conStr).connect()
    return connection


def executeQuery(query: str, host: str, database: str, useWindowsAuthentication: bool) -> pd.DataFrame:
    """
    Executes a SQL query.
    INPUT:
        `query`: a string
        `host`: a string
        `databse`: a string
        `useWindowsAuthentication`: a boolean
    OUTPUT:
        A pandas dataframe object.
    """
    databaseConnection = connectToDatabase(host, database, useWindowsAuthentication=useWindowsAuthentication)
    queryResult = pd.read_sql(query, databaseConnection)
    databaseConnection.close()
    return queryResult


# notes methods


def pull_metadata(note_version, id_type, note_type, sql_dir, cohort_dir, notes_dir, cohort, logger: Logger):
    logger.info(f"""Function arguments:
    `note_version`: {note_version}
    `id_type`:      {id_type}
    `note_type`:    {note_type}
    `sql_dir`:      {sql_dir}
    `cohort_dir`:   {cohort_dir}
    `notes_dir`:    {notes_dir}
    `cohort`:       {cohort}
    `logger`:       {logger}""")
    m = 'w'  # mode of the output file
    h = True  # header of the output file
    counter = 1  # used only for tracking/time estimation purposes
    if (id_type == 'PatientKey'):
        in_file = COHORT_FILE
    elif (id_type == 'EncounterCSN' or id_type == 'EncounterKey'):
        in_file = COHORT_FILE
    for df in pd.read_csv(os.path.join(cohort_dir, in_file), chunksize=1000, engine='python'):
        df = df[[id_type]]
        id_str = df[id_type].unique().tolist()
        logger.info(f"  This chunk of your input file containing patient or encounter ID's is of length {len(id_str):,}.")
        id_str = "','".join(str(x) for x in id_str)
        id_str = "'" + id_str + "'"
        query_file = note_type + '_metadata.sql'
        logger.info(f"Reading query file: {query_file}")
        fpath = os.path.join(sql_dir, query_file)
        with open(fpath, "r") as file:
            query = file.read()
        query = replace_sql_query(query, "XXXXX", id_str)
        query = replace_sql_query(query, "{PYTHON_VARIABLE: SQL_ENCOUNTER_EFFECTIVE_DATE_START}", SQL_ENCOUNTER_EFFECTIVE_DATE_START)
        query = replace_sql_query(query, "{PYTHON_VARIABLE: SQL_ENCOUNTER_EFFECTIVE_DATE_END}", SQL_ENCOUNTER_EFFECTIVE_DATE_END)

        logger.log(9, f"Using the following query:\n>>> Begin query >>>\n{query}\n<<< End query <<<")
        result = executeQuery(query=query,
                              host=host,
                              database=database_prod,
                              useWindowsAuthentication=useWindowsAuthentication)

        logger.info(f"The chunk query has {len(result):,} rows.")
        result = result.drop_duplicates()

        logger.info(f"After dropping duplicates, the chunk query has {len(result):,} rows.")
        if (note_type == 'note' and note_version == 'last'):  # keep only the last version of the note
            result = result.sort_values(by=['NoteKey', 'ContactNumber'], ascending=[True, False])
            result = result.drop_duplicates(['NoteKey'])
        result_file = cohort + '_' + note_type + '_metadata.csv'
        result.to_csv(os.path.join(notes_dir, result_file), index=False, mode=m, header=h)
        m = 'a'
        h = False
        logger.info(f'Completed chunk {counter}')
        counter = counter + 1
    return


def split_metadata(note_type, notes_dir, cohort, logger: Logger):
    in_file = cohort + '_' + note_type + '_metadata.csv'
    file_count = 1
    for df in pd.read_csv(os.path.join(notes_dir, in_file), chunksize=1000000):
        out_file = cohort + '_' + note_type + '_metadata_' + str(file_count) + '.csv'

        logger.info(f"Working on file {file_count}.")
        # ensure that all ID columns are integers
        columns = df.columns
        for c in ['NoteKey', 'NoteID', 'LinkageNoteID', 'OrderKey', 'OrderID', 'PatientKey', 'EncounterKey', 'EncounterCSN', 'AuthoringProviderKey', 'CosignProviderKey', 'OrderingProviderKey', 'AuthorizingProviderKey']:
            if (c in columns):
                # fill missing values with 0 since 0 is never used for these IDs in reality
                df[c] = df[c].fillna(0)
                # use string instead of int type since some fields, such as EncounterCSN, are bigger than what python considers an integer
                df[c] = df.apply(lambda row: str(row[c]).split('.0')[0], axis=1)
        df.to_csv(os.path.join(notes_dir, out_file), index=False)
        file_count = file_count + 1
    return


def pull_text(item, note_type, note_id, sql_dir, notes_dir, dir_final, logger: Logger):
    logger.info(f"""Processing item "{item}".""")
    # Pull text
    header = True  # header of the output file
    mode = 'w'  # mode of the output file
    for notes in pd.read_csv(os.path.join(notes_dir, item), chunksize=10000, engine='python'):
        note_list = notes[note_id].unique().tolist()
        note_list = ",".join(str(int(x)) for x in note_list)
        query_file = note_type + '_text.sql'
        query_file = os.path.join(sql_dir, query_file)
        fpath = os.path.join(sql_dir, query_file)
        with open(fpath, "r") as file:
            query = file.read()
        query = replace_sql_query(query, "XXXXX", note_list)
        result = executeQuery(query=query,
                              host=host,
                              database=database_notes,
                              useWindowsAuthentication=useWindowsAuthentication)
        # ensure that all ID columns are integers
        columns = result.columns
        for c in ['LinkageNoteID', 'OrderKey']:
            if (c in columns):
                # fill missing values with 0 since 0 is never used for these IDs in reality
                result[c] = result[c].fillna(0)
                # use string instead of int type since some fields, such as EncounterCSN, are bigger than what python considers an integer
                result[c] = result.apply(lambda row: str(row[c]).split('.0')[0], axis=1)
        file_count = item.split('.csv')[0].split('_')[-1]
        file_out = note_type + '_' + str(file_count) + '.tsv'
        result.to_csv(os.path.join(dir_final, file_out), header=header, mode=mode, sep='\t', encoding="UTF-8", index=False)
        header = False
        mode = 'a'
    return


def pull_text_in_parallel(note_type, sql_dir, notes_dir, cohort, logger: Logger):
    # Prepare for text pull
    if (note_type == 'note'):
        dir_final = os.path.join(notes_dir, cohort + '_note')
        if (not os.path.exists(dir_final)):
            os.makedirs(dir_final)
        note_id = 'LinkageNoteID'
    elif (note_type == 'order_narrative'):
        dir_final = os.path.join(notes_dir, cohort + '_order_narrative')
        if (not os.path.exists(dir_final)):
            os.makedirs(dir_final)
        note_id = 'OrderKey'
    elif (note_type == 'order_impression'):
        dir_final = os.path.join(notes_dir, cohort + '_order_impression')
        if (not os.path.exists(dir_final)):
            os.makedirs(dir_final)
        note_id = 'OrderKey'
    elif (note_type == 'order_result_comment'):
        dir_final = os.path.join(notes_dir, cohort + '_order_result_comment')
        if (not os.path.exists(dir_final)):
            os.makedirs(dir_final)
        note_id = 'OrderKey'
    # identify all metadata files for specific note type
    pattern = cohort + '_' + note_type + '_metadata_.*.csv'  # hot fix 2022-05-10
    items = [f for f in os.listdir(notes_dir) if re.match(pattern, f)]
    logger.info(f"""This is the list of items that matched the criteria for processing: {items}.""")
    # start pullng text. By default, number of parallel threads is set to 4.
    with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:
        result_futures = {executor.submit(pull_text, item, note_type, note_id, sql_dir, notes_dir, dir_final, logger): item for item in items}
        for future in concurrent.futures.as_completed(result_futures):
            item = result_futures[future]
            try:
                _ = future.result()
                logger.info(f"""Completed item "{item}".""")
            except Exception as e:
                logger.info(f"""An exception was generated by item "{item}": "{e}".""")
    return


def combine_order_metadata(notes_dir, cohort, logger: logging.Logger):
    ORDER_TYPE_LIST = ['order_impression',
                       'order_narrative',
                       'order_result_comment']
    LIST_OF_COLUMNS_TO_INCLUDE = ['OrderKey',
                                  'Line',
                                  'OrderID',
                                  'OrderPlacedDatetime',
                                  'OrderResultDatetime',
                                  'PatientKey',
                                  'MRN_GNV',
                                  'MRN_JAX',
                                  'NoteType',
                                  'EncounterDate',
                                  'EncounterKey',
                                  'EncounterCSN',
                                  'OrderingProviderKey',
                                  'OrderingProviderType',
                                  'OrderingProviderSpecialty',
                                  'AuthorizingProviderKey',
                                  'AuthorizingProviderType',
                                  'AuthorizingProviderSpecialty']
    HEADER = True
    MODE = 'w'
    CHUNKSIZE = 100000
    for orderType in ORDER_TYPE_LIST:
        file = cohort + '_' + orderType + '_metadata.csv'
        fpath = os.path.join(notes_dir, file)
        logger.info(f"""  Checking if file exists: "{fpath}".""")
        fileExists = os.path.exists(fpath)
        if fileExists:
            logger.info(f"""    File exists.""")
            chunkGenerator1 = pd.read_csv(fpath, chunksize=CHUNKSIZE)
            chunkGenerator2 = pd.read_csv(fpath, chunksize=CHUNKSIZE)
            logger.info("    Counting the number of chunks in the file.")
            it1Total = sum([1 for _ in chunkGenerator1])
            logger.info("    Counting the number of chunks in the file - done.")
            for it1, dfChunk in enumerate(chunkGenerator2, start=1):
                dfChunk = pd.DataFrame(dfChunk)  # For type hinting
                logger.info(f"""  ..  Working on chunk {it1:,} of {it1Total}.""")
                if (orderType in ['order_narrative', 'order_impression']):
                    dfChunk['Line'] = ''
                dfChunk = dfChunk[LIST_OF_COLUMNS_TO_INCLUDE]
                savePath = os.path.join(notes_dir, '{}_order_metadata.csv'.format(cohort))
                logger.info(f"""  ..  Saving chunk to "{savePath}".""")
                dfChunk.to_csv(path_or_buf=savePath,
                               index=False,
                               header=HEADER,
                               mode=MODE)
                HEADER = False
                MODE = "a"
        else:
            logger.info(f"""    Warning: File does not exist.""")  # Not sure if this merits a warning.

    # Final removal of duplicates
    logger.info("Performing final removal of duplicates.")
    df = pd.read_csv(savePath).drop_duplicates()
    df.to_csv(savePath, index=False)
    logger.info("Performing final removal of duplicates - done.")
    return


def create_encounters(notes_dir, logger: logging.Logger):
    CHUNKSIZE = 100000
    HEADER = True
    MODE = "w"
    fileList = ['{}_order_narrative_metadata.csv'.format(cohort),
                '{}_order_impression_metadata.csv'.format(cohort),
                '{}_order_result_comment_metadata.csv'.format(cohort),
                '{}_note_metadata.csv'.format(cohort)]
    it1Total = sum([1 for _ in fileList])
    for it1, file in enumerate(fileList, start=1):
        logger.info(f"""  Working on file {it1:,} of {it1Total:,}: "{file}".""")
        chunkGenerator1 = pd.read_csv(os.path.join(notes_dir, file), chunksize=CHUNKSIZE)
        chunkGenerator2 = pd.read_csv(os.path.join(notes_dir, file), chunksize=CHUNKSIZE)
        logger.info("    Counting the number of chunks in the file.")
        it2Total = sum([1 for _ in chunkGenerator1])
        logger.info(f"    Counting the number of chunks in the file - done. There are {it2Total:,} chunks.")
        for it2, dfChunk in enumerate(chunkGenerator2,  start=1):
            dfChunk = pd.DataFrame(dfChunk)  # For type hinting
            logger.info(f"""    Working on chunk {it2:,} of {it2Total:,}.""")
            df = dfChunk[['EncounterCSN']].drop_duplicates()
            savePath = os.path.join(notes_dir, 'encounters.csv')
            logger.info(f"""    Saving chunk to "{choosePathToLog(Path(savePath), rootDirectory)}".""")
            df.to_csv(path_or_buf=savePath,
                      index=False,
                      header=HEADER,
                      mode=MODE)  # TODO This is an intermediate file that might be eligible for removal.
            logger.info(f"""    Saving chunk to "{choosePathToLog(Path(savePath), rootDirectory)}" - done.""")
            HEADER = False
            MODE = "a"
        logger.info(f"""  Working on file {it1:,} of {it1Total:,}: "{file}" - done.""")

    # Final removal of duplicates
    logger.info("Performing final removal of duplicates.")
    df = pd.read_csv(savePath).drop_duplicates()
    df.to_csv(savePath, index=False)
    logger.info("Performing final removal of duplicates - done.")
    return


def create_provider_metadata(notes_dir, cohort, logger: logging.Logger):
    """
    Creates the provider metadata table. It does so by reading through all the metadata files in chunks and keeping the unique rows.
    """
    savePath = os.path.join(notes_dir, 'provider_metadata.csv')
    CHUNKSIZE = 100000
    HEADER = True
    MODE = "w"
    listOfClinicalTextTypes = ['note',
                               'order_impression',
                               'order_narrative',
                               'order_result_comment']
    it1Total = len(listOfClinicalTextTypes)
    for it1, file1 in enumerate(listOfClinicalTextTypes, start=1):
        logger.info(f"""  Work on file {it1:,} of {it1Total:,}: "{file1}".""")
        file = cohort + '_' + file1 + '_metadata.csv'
        fpath = os.path.join(notes_dir, file)
        logger.info(f"""  Checking if file exists: "{fpath}".""")
        fileExists = os.path.exists(fpath)
        if fileExists:
            logger.info(f"""    File exists.""")
            chunkGenerator1 = pd.read_csv(fpath, chunksize=CHUNKSIZE)
            chunkGenerator2 = pd.read_csv(fpath, chunksize=CHUNKSIZE)
            logger.info("    Counting the number of chunks in the file.")
            it2Total = sum([1 for _ in chunkGenerator1])
            logger.info(f"    Counting the number of chunks in the file - done. There are {it2Total:,} chunks.")
            for it2, dfChunk in enumerate(chunkGenerator2, start=1):
                logger.info(f"""    Working on chunk {it2:,} of {it2Total:,}.""")
                dfChunk = pd.DataFrame(dfChunk)  # For type hinting
                if (file1 in ['order_narrative', 'order_impression', 'order_result_comment']):
                    df1 = dfChunk[['OrderingProviderKey', 'OrderingProviderType', 'OrderingProviderSpecialty']]
                    df1 = df1.rename(columns={"OrderingProviderKey": "ProviderKey", "OrderingProviderType": "ProviderType", "OrderingProviderSpecialty": "ProviderSpecialty"})
                    df2 = dfChunk[['AuthorizingProviderKey', 'AuthorizingProviderType', 'AuthorizingProviderSpecialty']]
                    df2 = df2.rename(columns={"AuthorizingProviderKey": "ProviderKey", "AuthorizingProviderType": "ProviderType", "AuthorizingProviderSpecialty": "ProviderSpecialty"})
                elif (file1 in ['note']):
                    df1 = dfChunk[['AuthoringProviderKey', 'AuthoringProviderType', 'AuthoringProviderSpecialty']]
                    df1 = df1.rename(columns={"AuthoringProviderKey": "ProviderKey", "AuthoringProviderType": "ProviderType", "AuthoringProviderSpecialty": "ProviderSpecialty"})
                    df2 = dfChunk[['CosignProviderKey', 'CosignProviderType', 'CosignProviderSpecialty']]
                    df2 = df2.rename(columns={"CosignProviderKey": "ProviderKey", "CosignProviderType": "ProviderType", "CosignProviderSpecialty": "ProviderSpecialty"})
                    df = pd.concat([df1, df2])
                    df = df.drop_duplicates()
                    df.to_csv(path_or_buf=savePath,
                              index=False,
                              header=HEADER,
                              mode=MODE)
                    HEADER = False
                    MODE = "a"
        else:
            logger.info(f"""    Warning: File does not exist.""")  # Not sure if this merits a warning.

    # Final removal of duplicates
    logger.info("Performing final removal of duplicates.")
    df = pd.read_csv(savePath).drop_duplicates()
    df.to_csv(savePath, index=False)
    logger.info("Performing final removal of duplicates - done.")
    return


def generate_map(deid_mode, in_dir, map_dir, concept, cohort, logger: logging.Logger):
    """
    `concept` can be any of the following values:
        - patient
        - encounter
        - note
        - note_link
        - order
        - provider
    """
    # define variables
    if (concept == 'patient'):
        concept_cd = 'PAT'
        concept_id = 'deid_pat_id'
        file = os.path.join(in_dir, COHORT_FILE)  # NOTE
        concept_column = 'PatientKey'
    elif (concept == 'encounter'):
        concept_cd = 'ENC'
        concept_id = 'deid_enc_id'
        create_encounters(in_dir, logger=logger)
        file = os.path.join(in_dir, 'encounters.csv')
        concept_column = 'EncounterCSN'
    elif (concept == 'note'):
        concept_cd = 'NOTE'
        concept_id = 'deid_note_id'
        file = os.path.join(in_dir, '{}_note_metadata.csv'.format(cohort))
        concept_column = 'NoteKey'
    elif (concept == 'note_link'):
        concept_cd = 'LINK_NOTE'
        concept_id = 'deid_link_note_id'
        file = os.path.join(in_dir, '{}_note_metadata.csv'.format(cohort))
        concept_column = 'LinkageNoteID'
    elif (concept == 'order'):
        concept_cd = 'ORDER'
        concept_id = 'deid_order_id'
        combine_order_metadata(in_dir, cohort, logger=logger)
        file = os.path.join(in_dir, '{}_order_metadata.csv'.format(cohort))
        concept_column = 'OrderKey'
    elif (concept == 'provider'):
        concept_cd = 'PROV'
        concept_id = 'deid_provider_id'
        create_provider_metadata(in_dir, cohort, logger=logger)
        file = os.path.join(in_dir, 'provider_metadata.csv')
        concept_column = 'ProviderKey'
    else:
        logger.error(f"""Nonexisting concept in `generate_map` method: "{concept}".""")

    # Count the number of chunks in the file
    CHUNKSIZE = 10000
    chunkGenerator1 = pd.read_csv(file, chunksize=CHUNKSIZE)
    chunkGenerator2 = pd.read_csv(file, chunksize=CHUNKSIZE)
    logger.info("Counting the number of chunks in the file.")
    it1Total = sum([1 for _ in chunkGenerator1])
    logger.info(f"""Counting the number of chunks in the file - done. The are {it1Total:,} chunks.""")
    INDEX_START = 1
    HEADER = True
    MODE = "w"
    for it1, ids in enumerate(chunkGenerator2, start=1):
        logger.info(f"""  Working on chunk {it1:,} of {it1Total:,}.""")
        ids = pd.DataFrame(ids)  # For type hinting
        ids = ids[[concept_column]].drop_duplicates()
        ids.drop_duplicates(inplace=True)
        index_end = len(ids) + INDEX_START
        ids.index = range(INDEX_START, index_end)
        ids_map = ids.reset_index()
        ids_map['deid_num'] = ids_map.index
        # Assign de-identification value of "0" for unknown input (e.g., provider key < 0)
        ids_map[concept_id] = ids_map.apply(lambda row: (str(irb) + '_' + concept_cd + '_0') if (int(row[concept_column]) < 0) else (str(irb) + '_' + concept_cd + '_' + str(int(row['deid_num']))), axis=1)
        if (deid_mode == 'deid' and concept == 'patient'):
            ids_map['date_shift'] = ids_map.apply(lambda row: random.randint(-30, 30), axis=1)
        out_file = 'map_{}.csv'.format(concept)
        INDEX_START = max(ids_map["index"]) + 1
        ids_map = ids_map.drop(['index'], axis=1)
        savePath = Path(os.path.join(map_dir, out_file))
        logger.info(f"""  Saving chunk to "{choosePathToLog(savePath, rootDirectory)}".""")
        ids_map.to_csv(savePath, index=False, header=HEADER, mode=MODE)
        logger.info(f"""  Saving chunk to "{choosePathToLog(savePath, rootDirectory)}" - done.""")
        HEADER = False
        MODE = "a"
    return


def deid_metadata(deid_mode, note_type, map_dir, notes_dir, disclosure_dir):
    # define variables
    text_file = '{}_{}_metadata.csv'.format(cohort, note_type)
    if (note_type == 'note'):
        if (deid_mode == 'deid'):
            final_columns = ['deid_note_id', 'deid_link_note_id', 'ContactNumber', 'ContactDate_shifted', 'CreatedDatetime_shifted', 'ServiceDatetime_shifted', 'deid_pat_id', 'NoteType', 'InpatientNoteType', 'EncounterDate_shifted', 'deid_enc_id', 'deid_authoring_provider_id', 'AuthoringProviderType', 'AuthoringProviderSpecialty', 'deid_cosign_provider_id', 'CosignProviderType', 'CosignProviderSpecialty']
        elif (deid_mode == 'lds'):
            final_columns = ['deid_note_id', 'deid_link_note_id', 'ContactNumber', 'ContactDate', 'CreatedDatetime', 'ServiceDatetime', 'deid_pat_id', 'NoteType', 'InpatientNoteType', 'EncounterDate', 'deid_enc_id', 'deid_authoring_provider_id', 'AuthoringProviderType', 'AuthoringProviderSpecialty', 'deid_cosign_provider_id', 'CosignProviderType', 'CosignProviderSpecialty']
        elif (deid_mode == 'phi'):
            final_columns = ['NoteID', 'LinkageNoteID', 'ContactNumber', 'ContactDate', 'CreatedDatetime', 'ServiceDatetime', 'MRN_GNV', 'MRN_JAX', 'NoteType', 'InpatientNoteType', 'EncounterDate', 'EncounterCSN', 'AuthoringProviderKey', 'AuthoringProviderType', 'AuthoringProviderSpecialty', 'CosignProviderKey', 'CosignProviderType', 'CosignProviderSpecialty']
    elif (note_type == 'order'):
        if (deid_mode == 'deid'):
            final_columns = ['deid_order_id', 'Line', 'OrderPlacedDatetime_shifted', 'OrderResultDatetime_shifted', 'deid_pat_id', 'NoteType', 'EncounterDate_shifted', 'deid_enc_id', 'deid_ordering_provider_id', 'OrderingProviderType', 'OrderingProviderSpecialty', 'deid_authorizing_provider_id', 'AuthorizingProviderType', 'AuthorizingProviderSpecialty']
        elif (deid_mode == 'lds'):
            final_columns = ['deid_order_id', 'Line', 'OrderPlacedDatetime', 'OrderResultDatetime', 'deid_pat_id', 'NoteType', 'EncounterDate', 'deid_enc_id', 'deid_ordering_provider_id', 'OrderingProviderType', 'OrderingProviderSpecialty', 'deid_authorizing_provider_id', 'AuthorizingProviderType', 'AuthorizingProviderSpecialty']
        elif (deid_mode == 'phi'):
            final_columns = ['OrderID', 'Line', 'OrderPlacedDatetime', 'OrderResultDatetime', 'MRN_GNV', 'MRN_JAX', 'NoteType', 'EncounterDate', 'EncounterCSN', 'OrderingProviderKey', 'OrderingProviderType', 'OrderingProviderSpecialty', 'AuthorizingProviderKey', 'AuthorizingProviderType', 'AuthorizingProviderSpecialty']
    # import mappings
    map_pat = pd.read_csv(os.path.join(map_dir, 'map_patient.csv'))
    if (deid_mode == 'deid'):
        map_pat = map_pat[['PatientKey', 'deid_pat_id', 'date_shift']]
    else:
        map_pat = map_pat[['PatientKey', 'deid_pat_id']]
    map_enc = pd.read_csv(os.path.join(map_dir, 'map_encounter.csv'))
    map_enc = map_enc[['EncounterCSN', 'deid_enc_id']]
    map_prov = pd.read_csv(os.path.join(map_dir, 'map_provider.csv'))
    map_prov = map_prov[['ProviderKey', 'deid_provider_id']]
    if (note_type == 'order'):
        map_order = pd.read_csv(os.path.join(map_dir, 'map_order.csv'))
        map_order = map_order[['OrderKey', 'deid_order_id']]
    elif (note_type == 'note'):
        map_note = pd.read_csv(os.path.join(map_dir, 'map_note.csv'))
        map_note = map_note[['NoteKey', 'deid_note_id']]
        map_linkage_note = pd.read_csv(os.path.join(map_dir, 'map_note_link.csv'))
        map_linkage_note = map_linkage_note[['LinkageNoteID', 'deid_link_note_id']]

    # merge file with mapping files
    m = 'w'
    h = True
    for df in pd.read_csv(os.path.join(notes_dir, text_file), chunksize=100000):
        df = pd.merge(df, map_pat, how='left', on='PatientKey')
        df = df[df['PatientKey'] > 0]
        df = pd.merge(df, map_enc, how='left', on='EncounterCSN')
        df = df[df['EncounterKey'] > 0]
        if (note_type == 'order'):
            df = pd.merge(df, map_order, how='left', on='OrderKey')
            df = pd.merge(df, map_prov, how='left', left_on='OrderingProviderKey', right_on='ProviderKey')
            df = df.rename(columns={"deid_provider_id": "deid_ordering_provider_id"})
            df = pd.merge(df, map_prov, how='left', left_on='AuthorizingProviderKey', right_on='ProviderKey')
            df = df.rename(columns={"deid_provider_id": "deid_authorizing_provider_id"})
        elif (note_type == 'note'):
            df = pd.merge(df, map_note, how='left', on='NoteKey')
            df = pd.merge(df, map_linkage_note, how='left', on='LinkageNoteID')
            df = pd.merge(df, map_prov, how='left', left_on='AuthoringProviderKey', right_on='ProviderKey')
            df = df.rename(columns={"deid_provider_id": "deid_authoring_provider_id"})
            df = pd.merge(df, map_prov, how='left', left_on='CosignProviderKey', right_on='ProviderKey')
            df = df.rename(columns={"deid_provider_id": "deid_cosign_provider_id"})
        if (deid_mode == 'deid'):
            if (note_type == 'note'):
                df['ContactDate'] = pd.to_datetime(df['ContactDate'])
                df['ContactDate'] = df['ContactDate'].fillna('0')
                df['CreatedDatetime'] = pd.to_datetime(df['CreatedDatetime'])
                df['CreatedDatetime'] = df['CreatedDatetime'].fillna('0')
                df['ServiceDatetime'] = pd.to_datetime(df['ServiceDatetime'])
                df['ServiceDatetime'] = df['ServiceDatetime'].fillna('0')
                df['EncounterDate'] = pd.to_datetime(df['EncounterDate'])
                df['EncounterDate'] = df['EncounterDate'].fillna('0')
                df['ContactDate_shifted'] = df.apply(lambda row: row['ContactDate'] + timedelta(days=row['date_shift']) if row['ContactDate'] != '0' else '', axis=1)
                df['CreatedDatetime_shifted'] = df.apply(lambda row: row['CreatedDatetime'] + timedelta(days=row['date_shift']) if row['CreatedDatetime'] != '0' else '', axis=1)
                df['ServiceDatetime_shifted'] = df.apply(lambda row: row['ServiceDatetime'] + timedelta(days=row['date_shift']) if row['ServiceDatetime'] != '0' else '', axis=1)
                df['EncounterDate_shifted'] = df.apply(lambda row: row['EncounterDate'] + timedelta(days=row['date_shift']) if row['EncounterDate'] != '0' else '', axis=1)
            elif (note_type == 'order'):
                df['OrderPlacedDatetime'] = pd.to_datetime(df['OrderPlacedDatetime'])
                df['OrderPlacedDatetime'] = df['OrderPlacedDatetime'].fillna('0')
                df['OrderResultDatetime'] = pd.to_datetime(df['OrderResultDatetime'])
                df['OrderResultDatetime'] = df['OrderResultDatetime'].fillna('0')
                df['EncounterDate'] = pd.to_datetime(df['EncounterDate'])
                df['EncounterDate'] = df['EncounterDate'].fillna('0')
                df['OrderPlacedDatetime_shifted'] = df.apply(lambda row: row['OrderPlacedDatetime'] + timedelta(days=row['date_shift']) if row['OrderPlacedDatetime'] != '0' else '', axis=1)
                df['OrderResultDatetime_shifted'] = df.apply(lambda row: row['OrderResultDatetime'] + timedelta(days=row['date_shift']) if row['OrderResultDatetime'] != '0' else '', axis=1)
                df['EncounterDate_shifted'] = df.apply(lambda row: row['EncounterDate'] + timedelta(days=row['date_shift']) if row['EncounterDate'] != '0' else '', axis=1)
        df = df[final_columns]
        df.to_csv(os.path.join(disclosure_dir, text_file), index=False, mode=m, header=h)
        m = 'a'
        h = False
    return


def deid_tsv_note(map_dir, notes_dir, logger: Logger):
    map_note_link = pd.read_csv(os.path.join(map_dir, 'map_note_link.csv'))
    for file_prefix in ['note']:
        # find all files with specified prefix in the name
        pattern = file_prefix + '_.*.tsv'
        in_dir = os.path.join(notes_dir, '{}_{}'.format(cohort, file_prefix))
        final_columns = ['deid_link_note_id', 'note_text']
        items = [f for f in os.listdir(in_dir) if re.match(pattern, f)]
        for file in items:
            logger.info(f"""Processing `file` "{file}".""")
            out_file = 'deid_{}'.format(file)
            m = 'w'
            h = True
            for df in pd.read_csv(os.path.join(in_dir, file), chunksize=100000, sep='\t'):
                df = pd.merge(df, map_note_link, how='left', on='LinkageNoteID')
                df = df[final_columns]
                df.to_csv(os.path.join(in_dir, out_file), index=False, sep='\t', mode=m, header=h)
                m = 'a'
                h = False
    return


def deid_tsv_order(map_dir, notes_dir, logger: Logger):
    map_order = pd.read_csv(os.path.join(map_dir, 'map_order.csv'))
    for file_prefix in ['order_narrative', 'order_impression', 'order_result_comment']:
        # find all files with specified prefix in the name
        pattern = file_prefix + '_.*.tsv'
        in_dir = os.path.join(notes_dir, '{}_{}'.format(cohort, file_prefix))
        if (file_prefix == 'order_result_comment'):
            final_columns = ['deid_order_id', 'LINE', 'note_text']
        else:
            final_columns = ['deid_order_id', 'note_text']
        items = [f for f in os.listdir(in_dir) if re.match(pattern, f)]
        for file in items:
            logger.info(f"""    Processing `file` "{file}".""")
            out_file = 'deid_{}'.format(file)
            m = 'w'
            h = True
            for df in pd.read_csv(os.path.join(in_dir, file), chunksize=100000, sep='\t'):
                df = pd.merge(df, map_order, how='left', on='OrderKey')
                df = df[final_columns]
                df.to_csv(os.path.join(in_dir, out_file), index=False, sep='\t', mode=m, header=h)
                m = 'a'
                h = False
    return


def format_metadata_files(note_type, notes_dir, disclosure_dir):
    # define variables
    text_file = '{}_{}_metadata.csv'.format(cohort, note_type)
    if (note_type == 'note'):
        final_columns = ['NoteID', 'LinkageNoteID', 'ContactNumber', 'ContactDate', 'CreatedDatetime', 'ServiceDatetime', 'MRN_GNV', 'MRN_JAX', 'NoteType', 'InpatientNoteType', 'EncounterDate', 'EncounterCSN', 'AuthoringProviderKey', 'AuthoringProviderType', 'AuthoringProviderSpecialty', 'CosignProviderKey', 'CosignProviderType', 'CosignProviderSpecialty']
    elif (note_type == 'order'):
        final_columns = ['OrderID', 'Line', 'OrderPlacedDatetime', 'OrderResultDatetime', 'MRN_GNV', 'MRN_JAX', 'NoteType', 'EncounterDate', 'EncounterCSN', 'OrderingProviderKey', 'OrderingProviderType', 'OrderingProviderSpecialty', 'AuthorizingProviderKey', 'AuthorizingProviderType', 'AuthorizingProviderSpecialty']
    # read and format the file
    m = 'w'
    h = True
    for df in pd.read_csv(os.path.join(notes_dir, text_file), chunksize=100000):
        df = df[final_columns]
        df.to_csv(os.path.join(disclosure_dir, text_file), index=False, mode=m, header=h)
        m = 'a'
        h = False
    return


def copy_tsv(cohort, notes_dir, disclosure_dir):
    for item in ['note', 'order_narrative', 'order_impression', 'order_result_comment']:
        dir = cohort + '_' + item
        if (os.path.exists(os.path.join(notes_dir, dir))):
            file_list = [f for f in os.listdir(os.path.join(notes_dir, dir))]
            for f in file_list:
                shutil.copy(os.path.join(os.path.join(notes_dir, dir, f)), os.path.join(os.path.join(disclosure_dir, f)))
    return


# Logging block
logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

logger = logging.getLogger(__name__)

fileHandler = logging.FileHandler(logpath)
fileHandler.setLevel(9)
fileHandler.setFormatter(logFormat)

streamHandler = logging.StreamHandler()
streamHandler.setLevel(LOG_LEVEL)
streamHandler.setFormatter(logFormat)

logger.addHandler(fileHandler)
logger.addHandler(streamHandler)

logger.setLevel(9)

if __name__ == "__main__":
    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:

    # Arguments: Script arguments
    `COHORT_NAME`: "{COHORT_NAME}"
    `COHORT_FILE`: "{COHORT_FILE}"
    `IRB_NUMBER`: "{IRB_NUMBER}"
    `ID_TYPE`: "{ID_TYPE}"
    `NOTE_VERSION`: "{NOTE_VERSION}"
    `DE_IDENTIFICATION_MODE`: "{DE_IDENTIFICATION_MODE}"
    `SQL_ENCOUNTER_EFFECTIVE_DATE_START`: "{SQL_ENCOUNTER_EFFECTIVE_DATE_START}"
    `SQL_ENCOUNTER_EFFECTIVE_DATE_END`: "{SQL_ENCOUNTER_EFFECTIVE_DATE_END}"

    # Arguments: SQL connection settings
    `USE_WINDOWS_AUTHENTICATION` : "{USE_WINDOWS_AUTHENTICATION}"
    `SERVER`                     : "{SERVER}"
    `DATABASE_PROD`              : "{DATABASE_PROD}"
    `DATABASE_NOTES`             : "{DATABASE_NOTES}"
    `USERDOMAIN`                 : "{USERDOMAIN}"
    `USERNAME`                   : "{USERNAME}"
    `UID`                        : "{UID}"
    `PWD`                        : censored

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"
    """)
    logger.info(f"""`base_dir` set to "{base_dir}".""")

    # pull metadata
    pull_metadata(note_version, id_type, 'note', sql_dir, data_dir, notes_dir, cohort, logger)
    pull_metadata(note_version, id_type, 'order_narrative', sql_dir, data_dir, notes_dir, cohort, logger)
    pull_metadata(note_version, id_type, 'order_impression', sql_dir, data_dir, notes_dir, cohort, logger)
    pull_metadata(note_version, id_type, 'order_result_comment', sql_dir, data_dir, notes_dir, cohort, logger)

    # split metadata into chunks, so that we can process data in chunks
    split_metadata('note', notes_dir, cohort, logger)
    split_metadata('order_narrative', notes_dir, cohort, logger)
    split_metadata('order_impression', notes_dir, cohort, logger)
    split_metadata('order_result_comment', notes_dir, cohort, logger)

    # pull text in parallel
    pull_text_in_parallel('note', sql_dir, notes_dir, cohort, logger)
    pull_text_in_parallel('order_narrative', sql_dir, notes_dir, cohort, logger)
    pull_text_in_parallel('order_impression', sql_dir, notes_dir, cohort, logger)
    pull_text_in_parallel('order_result_comment', sql_dir, notes_dir, cohort, logger)
    if (deid_mode == 'phi'):

        # copy note_metadata file to disclosure folder
        format_metadata_files('note', notes_dir, disclosure_dir)

        # combine all order metadata files and copy to discosure folder
        combine_order_metadata(notes_dir, cohort)
        format_metadata_files('order', notes_dir, disclosure_dir)

        # copy .tsv files with free text
        copy_tsv(cohort, notes_dir, disclosure_dir)
    else:
        # Generate mappings
        logger.info("Generating de-identification maps for PHI variables.")
        logger.info("  Generating map for encounter IDs.")
        generate_map(deid_mode, notes_dir, map_dir, 'encounter', cohort, logger=logger)
        logger.info("  Generating map for encounter IDs - done.")
        logger.info("  Generating map for note IDs.")
        generate_map(deid_mode, notes_dir, map_dir, 'note', cohort, logger=logger)
        logger.info("  Generating map for note IDs - done.")
        logger.info("""  Generating map for note linkage IDs.""")
        generate_map(deid_mode, notes_dir, map_dir, 'note_link', cohort, logger=logger)
        logger.info("  Generating map for note linkage IDs.")
        logger.info("  Generating map for order IDs.")
        generate_map(deid_mode, notes_dir, map_dir, 'order', cohort, logger=logger)
        logger.info("  Generating map for order IDs - done.")
        logger.info("  Generating map for patient IDs.")
        generate_map(deid_mode, data_dir, map_dir, 'patient', cohort, logger=logger)
        logger.info("  Generating map for patient IDs - done.")
        logger.info("  Generating map for provider IDs.")
        generate_map(deid_mode, notes_dir, map_dir, 'provider', cohort, logger=logger)
        logger.info("  Generating map for provider IDs - done.")
        logger.info("Generating de-identification maps for PHI variables - done.")

        # De-identify metadata.
        logger.info("De-identifying metadata.")
        for clinicalTextCategory in ["note", "order"]:
            logger.info(f"""  De-identifying metadata for clinical text of category "{clinicalTextCategory}".""")
            deid_metadata(deid_mode, clinicalTextCategory, map_dir, notes_dir, disclosure_dir)
            logger.info(f"""  De-identifying metadata for clinical text of category "{clinicalTextCategory}" - done.""")
        logger.info("De-identifying metadata - done.")

        # De-identify clinical text, i.e., de-identify ID in .tsv file(s)
        logger.info("""De-identifying clinical text of category "note".""")
        deid_tsv_note(map_dir, notes_dir, logger)
        logger.info("""De-identifying clinical text of category "note" - done.""")
        logger.info("""De-identifying clinical text of category "order".""")
        deid_tsv_order(map_dir, notes_dir, logger)
        logger.info("""De-identifying clinical text of category "order" - done.""")

    # Output location summary
    logger.info(f"""Script output is located in the following directory: "{runOutputDir.absolute().relative_to(rootDirectory)}".""")

    # End script
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")



"""
Gather files in a central location for the honest broker to review

Define input files paths
Define input directory paths
Define destination directory path

create compression archive in destination directory path
for directory path in input directory paths
    move all contents to compression archive
for file in input file paths
    move file to compression archive
"""

from pathlib import Path


def gatherFiles(INPUT_FILE_PATHS,
                INPUT_DIRECTORY_PATHS,
                logger,
                ROOT_DIRECTORY,
                rootDirectory):
    """
    """
    functionName = __name__.split(".")[-1]
    logger.info(f"""Begin running "{functionName}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Function arguments:

    # Arguments
    ``: "{""}"
    """)

    # List all input files
    inputFilePaths = [Path(string) for string in INPUT_FILE_PATHS]
    inputDirectoryPaths = [Path(string) for string in INPUT_DIRECTORY_PATHS]

    # TODO Create compression archive

    # TODO Add all input files to compression archive
    _ = inputFilePaths
    _ = inputDirectoryPaths

    # TODO Move compression archive to destination directory

    print("This is not implemented. Move files manually. See script arguments for what files need to be moved.")



"""
Function(s) to modify clinical text output TSV to the right format for processing in the SDOH pipeline.
"""

import logging
from pathlib import Path
from typing import (Literal,
                    Union)
# Third-party packages
import pandas as pd


def modifyTSV(fpath: Union[str, Path],
              toDirectory: Union[str, Path],
              idColumnName: str = "LinkageNoteID"):
    """
    Modifies notes text TSV files in preparation for analysis in the SDOH pipeline.
    Modifications:
        - Adds two dummy columns
            - "CNTCT_NUM" with value "0"
            - "NOTE_KEY" with value "123456789"
        - Re-orders the columns
        - Renames the clinical text ID column from `idColumnName` to "NOTE_ENCNTR_KEY"
    """
    runOutputDir = Path(toDirectory)

    # Read file
    df = pd.read_csv(fpath, delimiter="\t")

    # Set dummy values
    df["CNTCT_NUM"] = 0
    df["NOTE_KEY"] = 123456789

    # Re-order columns
    df = df[[idColumnName,
             "NOTE_KEY",
             "CNTCT_NUM",
             "note_text"]]
    
    # Rename columns
    df = df.rename(columns={idColumnName: "NOTE_ENCNTR_KEY"})

    # Save modified TSV file
    fname = fpath.name
    savePath = runOutputDir.joinpath(fname)
    df.to_csv(savePath, index=False, sep="\t")


def wrapModify(fpath: Path,
               textType: Literal["note", "order"],
               toDirectory: Union[Path, str],
               logger: logging.Logger):
    """
    A wrapper function for `modifyTSV` to be called in parallel work.
    """
    logger.info(f"""  Working on path "{fpath}".""")
    if textType == "note":
        if "deid" in fpath.name:
            modifyTSV(fpath=fpath,
                    toDirectory=toDirectory,
                    idColumnName="deid_link_note_id")
        else:
            modifyTSV(fpath=fpath,
                    toDirectory=toDirectory,
                    idColumnName="LinkageNoteID")
    elif textType == "order":
        if "deid" in fpath.name:
            modifyTSV(fpath=fpath,
                    toDirectory=toDirectory,
                    idColumnName="deid_order_id")
        else:
            modifyTSV(fpath=fpath,
                    toDirectory=toDirectory,
                    idColumnName="OrderKey")
    else:
        raise Exception(f"""Expected `textType` to be one of "note" or "order", but got "{textType}" instead.""")
    logger.info(f"""  Working on path "{fpath}" - done.""")



"""
This is the modular implementation of "freeText.py".

TODO: Convert SCRIPT_ARGUMENTS to functionArguments in the functions below. Script arguments are then fed into `main` which are in turn passed to the functions.
"""

import concurrent.futures
import logging
import os
import pandas as pd
import random
import re
import shutil
import sqlalchemy as sa
from datetime import timedelta
from drapi.drapi import makeDirPath, replace_sql_query
from pathlib import Path


# Variables: Path construction
base_dir = str(Path(__file__).parent.parent.absolute())
data_dir = os.path.join(base_dir, 'data')
sql_dir = os.path.join(base_dir, 'sql')
notes_dir = os.path.join(data_dir, "Output", 'free_text')  # all notes related files are saved in 'notes' subdirectory of 'data' directory
map_dir = os.path.join(data_dir, "Output", 'mapping')  # mappings are saved in 'mapping' subdirectory of 'data' folder.
disclosure_dir = os.path.join(data_dir, "Output", 'disclosure')
for dir in [data_dir, sql_dir, notes_dir, map_dir, disclosure_dir]:
    makeDirPath(dir)

# Variables: SQL connection settings
host = 'DWSRSRCH01.shands.ufl.edu'
database_prod = 'DWS_PROD'
database_notes = 'DWS_NOTES'

# Variables: Script settings
cohort = COHORT_NAME
id_type = ID_TYPE
note_version = NOTE_VERSION
irb = IRB_NUMBER
deid_mode = DE_IDENTIFICATION_MODE


def db_connect(host, database):
    connstr = f"mssql+pymssql://{UID}:{PWD}@{host}/{database}"
    engine = sa.create_engine(connstr)
    return engine


def db_close(db_conn):
    db_conn.dispose()


def db_query(query, db_connection):
    return pd.read_sql(query, db_connection)


def db_execute_read_query(query, host, database):
    db_conn = db_connect(host, database)
    result_df = db_query(query, db_conn)
    db_close(db_conn)
    return result_df


def read_sql_file(file):
    with open(file) as f:
        query = f.read()
    f.close()
    return query


# notes methods


def pull_metadata(note_version, id_type, note_type, sql_dir, cohort_dir, notes_dir, cohort):
    message = f"""Working on `note_version`: {note_version}
    `id_type`: {id_type}
    `note_type`: {note_type}
    `sql_dir`: {sql_dir}
    `cohort_dir`: {cohort_dir}
    `notes_dir`: {notes_dir}
    `cohort`: {cohort}"""
    logging.debug(message)
    m = 'w'  # mode of the output file
    h = True  # header of the output file
    counter = 1  # used only for tracking/time estimation purposes
    if (id_type == 'PatientKey'):
        in_file = COHORT_FILE
    elif (id_type == 'EncounterCSN' or id_type == 'EncounterKey'):
        in_file = COHORT_FILE
    for df in pd.read_csv(os.path.join(cohort_dir, in_file), chunksize=1000, engine='python'):
        df = df[[id_type]]
        id_str = df[id_type].unique().tolist()
        message = f"This chunk of your input file containing patient or encounter ID's is of length {len(id_str)}."
        logging.debug(message)
        id_str = "','".join(str(x) for x in id_str)
        id_str = "'" + id_str + "'"
        query_file = note_type + '_metadata.sql'
        logging.debug(f"Reading query file: {query_file}")
        query = read_sql_file(os.path.join(sql_dir, query_file))
        query = replace_sql_query(query, "XXXXX", id_str)
        query = replace_sql_query(query, "{PYTHON_VARIABLE: SQL_ENCOUNTER_EFFECTIVE_DATE_START}", SQL_ENCOUNTER_EFFECTIVE_DATE_START)
        query = replace_sql_query(query, "{PYTHON_VARIABLE: SQL_ENCOUNTER_EFFECTIVE_DATE_END}", SQL_ENCOUNTER_EFFECTIVE_DATE_END)
        message = f"Using the following query:\n>>> Begin query >>>\n{query}\n<<< End query <<<"
        logging.log(9, message)
        result = db_execute_read_query(query, host, database_prod)
        message = f"The chunk query has {len(result)} rows."
        logging.debug(message)
        result = result.drop_duplicates()
        message = f"After dropping duplicates, the chunk query has {len(result)} rows."
        logging.debug(message)
        if (note_type == 'note' and note_version == 'last'):  # keep only the last version of the note
            result = result.sort_values(by=['NoteKey', 'ContactNumber'], ascending=[True, False])
            result = result.drop_duplicates(['NoteKey'])
        result_file = cohort + '_' + note_type + '_metadata.csv'
        result.to_csv(os.path.join(notes_dir, result_file), index=False, mode=m, header=h)
        m = 'a'
        h = False
        logging.info(f'Completed chunk {counter}')
        counter = counter + 1
    return


def split_metadata(note_type, notes_dir, cohort):
    in_file = cohort + '_' + note_type + '_metadata.csv'
    file_count = 1
    for df in pd.read_csv(os.path.join(notes_dir, in_file), chunksize=1000000):
        out_file = cohort + '_' + note_type + '_metadata_' + str(file_count) + '.csv'
        logging.debug(f"Working on file {file_count}.")
        # ensure that all ID columns are integers
        columns = df.columns
        for c in ['NoteKey', 'NoteID', 'LinkageNoteID', 'OrderKey', 'OrderID', 'PatientKey', 'EncounterKey', 'EncounterCSN', 'AuthoringProviderKey', 'CosignProviderKey', 'OrderingProviderKey', 'AuthorizingProviderKey']:
            if (c in columns):
                # fill missing values with 0 since 0 is never used for these IDs in reality
                df[c] = df[c].fillna(0)
                # use string instead of int type since some fields, such as EncounterCSN, are bigger than what python considers an integer
                df[c] = df.apply(lambda row: str(row[c]).split('.0')[0], axis=1)
        df.to_csv(os.path.join(notes_dir, out_file), index=False)
        file_count = file_count + 1
    return


def pull_text(item, note_type, note_id, sql_dir, notes_dir, dir_final):
    print('Processing {}'.format(item))
    # Pull text
    header = True  # header of the output file
    mode = 'w'  # mode of the output file
    for notes in pd.read_csv(os.path.join(notes_dir, item), chunksize=10000, engine='python'):
        note_list = notes[note_id].unique().tolist()
        note_list = ",".join(str(int(x)) for x in note_list)
        query_file = note_type + '_text.sql'
        query_file = os.path.join(sql_dir, query_file)
        query = read_sql_file(query_file)
        query = replace_sql_query(query, "XXXXX", note_list)
        result = db_execute_read_query(query, host, database_notes)
        # ensure that all ID columns are integers
        columns = result.columns
        for c in ['LinkageNoteID', 'OrderKey']:
            if (c in columns):
                # fill missing values with 0 since 0 is never used for these IDs in reality
                result[c] = result[c].fillna(0)
                # use string instead of int type since some fields, such as EncounterCSN, are bigger than what python considers an integer
                result[c] = result.apply(lambda row: str(row[c]).split('.0')[0], axis=1)
        file_count = item.split('.csv')[0].split('_')[-1]
        file_out = note_type + '_' + str(file_count) + '.tsv'
        result.to_csv(os.path.join(dir_final, file_out), header=header, mode=mode, sep='\t', encoding="UTF-8", index=False)
        header = False
        mode = 'a'
    return


def pull_text_in_parallel(note_type, sql_dir, notes_dir, cohort):
    # Prepare for text pull
    if (note_type == 'note'):
        dir_final = os.path.join(notes_dir, cohort + '_note')
        if (not os.path.exists(dir_final)):
            os.makedirs(dir_final)
        note_id = 'LinkageNoteID'
    elif (note_type == 'order_narrative'):
        dir_final = os.path.join(notes_dir, cohort + '_order_narrative')
        if (not os.path.exists(dir_final)):
            os.makedirs(dir_final)
        note_id = 'OrderKey'
    elif (note_type == 'order_impression'):
        dir_final = os.path.join(notes_dir, cohort + '_order_impression')
        if (not os.path.exists(dir_final)):
            os.makedirs(dir_final)
        note_id = 'OrderKey'
    elif (note_type == 'order_result_comment'):
        dir_final = os.path.join(notes_dir, cohort + '_order_result_comment')
        if (not os.path.exists(dir_final)):
            os.makedirs(dir_final)
        note_id = 'OrderKey'
    # identify all metadata files for specific note type
    pattern = cohort + '_' + note_type + '_metadata_.*.csv'  # hot fix 2022-05-10
    items = [f for f in os.listdir(notes_dir) if re.match(pattern, f)]
    print(items)
    # start pullng text. By default, number of parallel threads is set to 4.
    with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:
        result_futures = {executor.submit(pull_text, item, note_type, note_id, sql_dir, notes_dir, dir_final): item for item in items}
        for future in concurrent.futures.as_completed(result_futures):
            item = result_futures[future]
            try:
                _ = future.result()
                print('completed {}'.format(item))  # this is just to know that the item has been processed
            except Exception as e:
                print('{} generated an exception: {}'.format(item, e))  # this is used to get notification when there is an error, so something was not processed properly
    return


def combine_order_metadata(notes_dir, cohort):
    m = 'w'
    h = True
    for file1 in ['order_narrative', 'order_impression', 'order_result_comment']:
        file = cohort + '_' + file1 + '_metadata.csv'
        if os.path.exists(os.path.join(notes_dir, file)):
            df = pd.read_csv(os.path.join(notes_dir, file))
            if (file1 in ['order_narrative', 'order_impression']):
                df['Line'] = ''
            df = df[['OrderKey', 'Line', 'OrderID', 'OrderPlacedDatetime', 'OrderResultDatetime', 'PatientKey', 'MRN_GNV', 'MRN_JAX', 'NoteType', 'EncounterDate', 'EncounterKey', 'EncounterCSN', 'OrderingProviderKey', 'OrderingProviderType', 'OrderingProviderSpecialty', 'AuthorizingProviderKey', 'AuthorizingProviderType', 'AuthorizingProviderSpecialty']]
            df.to_csv(os.path.join(notes_dir, '{}_order_metadata.csv'.format(cohort)), index=False, mode=m, header=h)
            m = 'a'
            h = False
    return


def create_encounters(notes_dir):
    df = pd.DataFrame()
    for file in ['{}_order_narrative_metadata.csv'.format(cohort), '{}_order_impression_metadata.csv'.format(cohort), '{}_order_result_comment_metadata.csv'.format(cohort), '{}_note_metadata.csv'.format(cohort)]:
        for dfx in pd.read_csv(os.path.join(notes_dir, file), chunksize=100000):
            df1 = dfx[['EncounterCSN']].drop_duplicates()
            df = pd.concat([df, df1])
            df.drop_duplicates(inplace=True)
    df.to_csv(os.path.join(notes_dir, 'encounters.csv'), index=False)
    return


def create_provider_metadata(notes_dir, cohort):
    m = 'w'
    h = True
    df = pd.DataFrame()
    for file1 in ['order_narrative', 'order_impression', 'order_result_comment', 'note']:
        # for file in ['note']:
        file = cohort + '_' + file1 + '_metadata.csv'
        if os.path.exists(os.path.join(notes_dir, file)):
            for dfx in pd.read_csv(os.path.join(notes_dir, file), chunksize=100000):
                if (file1 in ['order_narrative', 'order_impression', 'order_result_comment']):
                    df1 = dfx[['OrderingProviderKey', 'OrderingProviderType', 'OrderingProviderSpecialty']]
                    df1 = df1.rename(columns={"OrderingProviderKey": "ProviderKey", "OrderingProviderType": "ProviderType", "OrderingProviderSpecialty": "ProviderSpecialty"})
                    df2 = dfx[['AuthorizingProviderKey', 'AuthorizingProviderType', 'AuthorizingProviderSpecialty']]
                    df2 = df2.rename(columns={"AuthorizingProviderKey": "ProviderKey", "AuthorizingProviderType": "ProviderType", "AuthorizingProviderSpecialty": "ProviderSpecialty"})
                elif (file1 in ['note']):
                    df1 = dfx[['AuthoringProviderKey', 'AuthoringProviderType', 'AuthoringProviderSpecialty']]
                    df1 = df1.rename(columns={"AuthoringProviderKey": "ProviderKey", "AuthoringProviderType": "ProviderType", "AuthoringProviderSpecialty": "ProviderSpecialty"})
                    df2 = dfx[['CosignProviderKey', 'CosignProviderType', 'CosignProviderSpecialty']]
                    df2 = df2.rename(columns={"CosignProviderKey": "ProviderKey", "CosignProviderType": "ProviderType", "CosignProviderSpecialty": "ProviderSpecialty"})
                df1 = pd.concat([df1, df2])
                df1.drop_duplicates(inplace=True)
                df = pd.concat([df, df1])
                df.drop_duplicates(inplace=True)
            df.to_csv(os.path.join(notes_dir, 'provider_metadata.csv'), index=False, mode=m, header=h)
            m = 'a'
            h = False
    return


def generate_map(deid_mode, in_dir, map_dir, concept, cohort):  # concept can be patient, encounter, note, note_link, order, or provider
    # define variables
    if (concept == 'patient'):
        concept_cd = 'PAT'
        concept_id = 'deid_pat_id'
        file = os.path.join(in_dir, 'cohort.csv')
        concept_column = 'PatientKey'
    elif (concept == 'encounter'):
        concept_cd = 'ENC'
        concept_id = 'deid_enc_id'
        create_encounters(in_dir)
        file = os.path.join(in_dir, 'encounters.csv')
        concept_column = 'EncounterCSN'
    elif (concept == 'note'):
        concept_cd = 'NOTE'
        concept_id = 'deid_note_id'
        file = os.path.join(in_dir, '{}_note_metadata.csv'.format(cohort))
        concept_column = 'NoteKey'
    elif (concept == 'note_link'):
        concept_cd = 'LINK_NOTE'
        concept_id = 'deid_link_note_id'
        file = os.path.join(in_dir, '{}_note_metadata.csv'.format(cohort))
        concept_column = 'LinkageNoteID'
    elif (concept == 'order'):
        concept_cd = 'ORDER'
        concept_id = 'deid_order_id'
        combine_order_metadata(in_dir, cohort)
        file = os.path.join(in_dir, '{}_order_metadata.csv'.format(cohort))
        concept_column = 'OrderKey'
    elif (concept == 'provider'):
        concept_cd = 'PROV'
        concept_id = 'deid_provider_id'
        create_provider_metadata(in_dir, cohort)
        file = os.path.join(in_dir, 'provider_metadata.csv')
        concept_column = 'ProviderKey'
    else:
        print('Nonexisting concept in generate_map method')

    ids_final = pd.DataFrame()
    for ids in pd.read_csv(file, chunksize=10000):
        ids = ids[[concept_column]].drop_duplicates()
        ids_final = pd.concat([ids_final, ids])
        ids_final.drop_duplicates(inplace=True)
    ids_map = ids_final.reset_index()
    ids_map['deid_num'] = ids_map.index + 1
    # assign deid value 0 for unknown input (e.g., provider key < 0)
    ids_map[concept_id] = ids_map.apply(lambda row: (str(irb) + '_' + concept_cd + '_0') if (int(row[concept_column]) < 0) else (str(irb) + '_' + concept_cd + '_' + str(int(row['deid_num']))), axis=1)
    if (deid_mode == 'deid' and concept == 'patient'):
        ids_map['date_shift'] = ids_map.apply(lambda row: random.randint(-30, 30), axis=1)
    out_file = 'map_{}.csv'.format(concept)
    ids_map = ids_map.drop(['index'], axis=1)
    ids_map.to_csv(os.path.join(map_dir, out_file), index=False)
    return


def deid_metadata(deid_mode, note_type, map_dir, notes_dir, disclosure_dir):
    # define variables
    text_file = '{}_{}_metadata.csv'.format(cohort, note_type)
    if (note_type == 'note'):
        if (deid_mode == 'deid'):
            final_columns = ['deid_note_id', 'deid_link_note_id', 'ContactNumber', 'ContactDate_shifted', 'CreatedDatetime_shifted', 'ServiceDatetime_shifted', 'deid_pat_id', 'NoteType', 'InpatientNoteType', 'EncounterDate_shifted', 'deid_enc_id', 'deid_authoring_provider_id', 'AuthoringProviderType', 'AuthoringProviderSpecialty', 'deid_cosign_provider_id', 'CosignProviderType', 'CosignProviderSpecialty']
        elif (deid_mode == 'lds'):
            final_columns = ['deid_note_id', 'deid_link_note_id', 'ContactNumber', 'ContactDate', 'CreatedDatetime', 'ServiceDatetime', 'deid_pat_id', 'NoteType', 'InpatientNoteType', 'EncounterDate', 'deid_enc_id', 'deid_authoring_provider_id', 'AuthoringProviderType', 'AuthoringProviderSpecialty', 'deid_cosign_provider_id', 'CosignProviderType', 'CosignProviderSpecialty']
        elif (deid_mode == 'phi'):
            final_columns = ['NoteID', 'LinkageNoteID', 'ContactNumber', 'ContactDate', 'CreatedDatetime', 'ServiceDatetime', 'MRN_GNV', 'MRN_JAX', 'NoteType', 'InpatientNoteType', 'EncounterDate', 'EncounterCSN', 'AuthoringProviderKey', 'AuthoringProviderType', 'AuthoringProviderSpecialty', 'CosignProviderKey', 'CosignProviderType', 'CosignProviderSpecialty']
    elif (note_type == 'order'):
        if (deid_mode == 'deid'):
            final_columns = ['deid_order_id', 'Line', 'OrderPlacedDatetime_shifted', 'OrderResultDatetime_shifted', 'deid_pat_id', 'NoteType', 'EncounterDate_shifted', 'deid_enc_id', 'deid_ordering_provider_id', 'OrderingProviderType', 'OrderingProviderSpecialty', 'deid_authorizing_provider_id', 'AuthorizingProviderType', 'AuthorizingProviderSpecialty']
        elif (deid_mode == 'lds'):
            final_columns = ['deid_order_id', 'Line', 'OrderPlacedDatetime', 'OrderResultDatetime', 'deid_pat_id', 'NoteType', 'EncounterDate', 'deid_enc_id', 'deid_ordering_provider_id', 'OrderingProviderType', 'OrderingProviderSpecialty', 'deid_authorizing_provider_id', 'AuthorizingProviderType', 'AuthorizingProviderSpecialty']
        elif (deid_mode == 'phi'):
            final_columns = ['OrderID', 'Line', 'OrderPlacedDatetime', 'OrderResultDatetime', 'MRN_GNV', 'MRN_JAX', 'NoteType', 'EncounterDate', 'EncounterCSN', 'OrderingProviderKey', 'OrderingProviderType', 'OrderingProviderSpecialty', 'AuthorizingProviderKey', 'AuthorizingProviderType', 'AuthorizingProviderSpecialty']
    # import mappings
    map_pat = pd.read_csv(os.path.join(map_dir, 'map_patient.csv'))
    if (deid_mode == 'deid'):
        map_pat = map_pat[['PatientKey', 'deid_pat_id', 'date_shift']]
    else:
        map_pat = map_pat[['PatientKey', 'deid_pat_id']]
    map_enc = pd.read_csv(os.path.join(map_dir, 'map_encounter.csv'))
    map_enc = map_enc[['EncounterCSN', 'deid_enc_id']]
    map_prov = pd.read_csv(os.path.join(map_dir, 'map_provider.csv'))
    map_prov = map_prov[['ProviderKey', 'deid_provider_id']]
    if (note_type == 'order'):
        map_order = pd.read_csv(os.path.join(map_dir, 'map_order.csv'))
        map_order = map_order[['OrderKey', 'deid_order_id']]
    elif (note_type == 'note'):
        map_note = pd.read_csv(os.path.join(map_dir, 'map_note.csv'))
        map_note = map_note[['NoteKey', 'deid_note_id']]
        map_linkage_note = pd.read_csv(os.path.join(map_dir, 'map_note_link.csv'))
        map_linkage_note = map_linkage_note[['LinkageNoteID', 'deid_link_note_id']]

    # merge file with mapping files
    m = 'w'
    h = True
    for df in pd.read_csv(os.path.join(notes_dir, text_file), chunksize=100000):
        df = pd.merge(df, map_pat, how='left', on='PatientKey')
        df = df[df['PatientKey'] > 0]
        df = pd.merge(df, map_enc, how='left', on='EncounterCSN')
        df = df[df['EncounterKey'] > 0]
        if (note_type == 'order'):
            df = pd.merge(df, map_order, how='left', on='OrderKey')
            df = pd.merge(df, map_prov, how='left', left_on='OrderingProviderKey', right_on='ProviderKey')
            df = df.rename(columns={"deid_provider_id": "deid_ordering_provider_id"})
            df = pd.merge(df, map_prov, how='left', left_on='AuthorizingProviderKey', right_on='ProviderKey')
            df = df.rename(columns={"deid_provider_id": "deid_authorizing_provider_id"})
        elif (note_type == 'note'):
            df = pd.merge(df, map_note, how='left', on='NoteKey')
            df = pd.merge(df, map_linkage_note, how='left', on='LinkageNoteID')
            df = pd.merge(df, map_prov, how='left', left_on='AuthoringProviderKey', right_on='ProviderKey')
            df = df.rename(columns={"deid_provider_id": "deid_authoring_provider_id"})
            df = pd.merge(df, map_prov, how='left', left_on='CosignProviderKey', right_on='ProviderKey')
            df = df.rename(columns={"deid_provider_id": "deid_cosign_provider_id"})
        if (deid_mode == 'deid'):
            if (note_type == 'note'):
                df['ContactDate'] = pd.to_datetime(df['ContactDate'])
                df['ContactDate'] = df['ContactDate'].fillna('0')
                df['CreatedDatetime'] = pd.to_datetime(df['CreatedDatetime'])
                df['CreatedDatetime'] = df['CreatedDatetime'].fillna('0')
                df['ServiceDatetime'] = pd.to_datetime(df['ServiceDatetime'])
                df['ServiceDatetime'] = df['ServiceDatetime'].fillna('0')
                df['EncounterDate'] = pd.to_datetime(df['EncounterDate'])
                df['EncounterDate'] = df['EncounterDate'].fillna('0')
                df['ContactDate_shifted'] = df.apply(lambda row: row['ContactDate'] + timedelta(days=row['date_shift']) if row['ContactDate'] != '0' else '', axis=1)
                df['CreatedDatetime_shifted'] = df.apply(lambda row: row['CreatedDatetime'] + timedelta(days=row['date_shift']) if row['CreatedDatetime'] != '0' else '', axis=1)
                df['ServiceDatetime_shifted'] = df.apply(lambda row: row['ServiceDatetime'] + timedelta(days=row['date_shift']) if row['ServiceDatetime'] != '0' else '', axis=1)
                df['EncounterDate_shifted'] = df.apply(lambda row: row['EncounterDate'] + timedelta(days=row['date_shift']) if row['EncounterDate'] != '0' else '', axis=1)
            elif (note_type == 'order'):
                df['OrderPlacedDatetime'] = pd.to_datetime(df['OrderPlacedDatetime'])
                df['OrderPlacedDatetime'] = df['OrderPlacedDatetime'].fillna('0')
                df['OrderResultDatetime'] = pd.to_datetime(df['OrderResultDatetime'])
                df['OrderResultDatetime'] = df['OrderResultDatetime'].fillna('0')
                df['EncounterDate'] = pd.to_datetime(df['EncounterDate'])
                df['EncounterDate'] = df['EncounterDate'].fillna('0')
                df['OrderPlacedDatetime_shifted'] = df.apply(lambda row: row['OrderPlacedDatetime'] + timedelta(days=row['date_shift']) if row['OrderPlacedDatetime'] != '0' else '', axis=1)
                df['OrderResultDatetime_shifted'] = df.apply(lambda row: row['OrderResultDatetime'] + timedelta(days=row['date_shift']) if row['OrderResultDatetime'] != '0' else '', axis=1)
                df['EncounterDate_shifted'] = df.apply(lambda row: row['EncounterDate'] + timedelta(days=row['date_shift']) if row['EncounterDate'] != '0' else '', axis=1)
        df = df[final_columns]
        df.to_csv(os.path.join(disclosure_dir, text_file), index=False, mode=m, header=h)
        m = 'a'
        h = False
    return


def deid_tsv_note(map_dir, notes_dir):
    map_note_link = pd.read_csv(os.path.join(map_dir, 'map_note_link.csv'))
    for file_prefix in ['note']:
        # find all files with specified prefix in the name
        pattern = file_prefix + '_.*.tsv'
        in_dir = os.path.join(notes_dir, '{}_{}'.format(cohort, file_prefix))
        final_columns = ['deid_link_note_id', 'note_text']
        items = [f for f in os.listdir(in_dir) if re.match(pattern, f)]
        for file in items:
            print('processing {}'.format(file))
            out_file = 'deid_{}'.format(file)
            m = 'w'
            h = True
            for df in pd.read_csv(os.path.join(in_dir, file), chunksize=100000, sep='\t'):
                df = pd.merge(df, map_note_link, how='left', on='LinkageNoteID')
                df = df[final_columns]
                df.to_csv(os.path.join(in_dir, out_file), index=False, sep='\t', mode=m, header=h)
                m = 'a'
                h = False
    return


def deid_tsv_order(map_dir, notes_dir):
    map_order = pd.read_csv(os.path.join(map_dir, 'map_order.csv'))
    for file_prefix in ['order_narrative', 'order_impression', 'order_result_comment']:
        # find all files with specified prefix in the name
        pattern = file_prefix + '_.*.tsv'
        in_dir = os.path.join(notes_dir, '{}_{}'.format(cohort, file_prefix))
        if (file_prefix == 'order_result_comment'):
            final_columns = ['deid_order_id', 'LINE', 'note_text']
        else:
            final_columns = ['deid_order_id', 'note_text']
        items = [f for f in os.listdir(in_dir) if re.match(pattern, f)]
        for file in items:
            print('processing {}'.format(file))
            out_file = 'deid_{}'.format(file)
            m = 'w'
            h = True
            for df in pd.read_csv(os.path.join(in_dir, file), chunksize=100000, sep='\t'):
                df = pd.merge(df, map_order, how='left', on='OrderKey')
                df = df[final_columns]
                df.to_csv(os.path.join(in_dir, out_file), index=False, sep='\t', mode=m, header=h)
                m = 'a'
                h = False
    return


def format_metadata_files(note_type, notes_dir, disclosure_dir):
    # define variables
    text_file = '{}_{}_metadata.csv'.format(cohort, note_type)
    if (note_type == 'note'):
        final_columns = ['NoteID', 'LinkageNoteID', 'ContactNumber', 'ContactDate', 'CreatedDatetime', 'ServiceDatetime', 'MRN_GNV', 'MRN_JAX', 'NoteType', 'InpatientNoteType', 'EncounterDate', 'EncounterCSN', 'AuthoringProviderKey', 'AuthoringProviderType', 'AuthoringProviderSpecialty', 'CosignProviderKey', 'CosignProviderType', 'CosignProviderSpecialty']
    elif (note_type == 'order'):
        final_columns = ['OrderID', 'Line', 'OrderPlacedDatetime', 'OrderResultDatetime', 'MRN_GNV', 'MRN_JAX', 'NoteType', 'EncounterDate', 'EncounterCSN', 'OrderingProviderKey', 'OrderingProviderType', 'OrderingProviderSpecialty', 'AuthorizingProviderKey', 'AuthorizingProviderType', 'AuthorizingProviderSpecialty']
    # read and format the file
    m = 'w'
    h = True
    for df in pd.read_csv(os.path.join(notes_dir, text_file), chunksize=100000):
        df = df[final_columns]
        df.to_csv(os.path.join(disclosure_dir, text_file), index=False, mode=m, header=h)
        m = 'a'
        h = False
    return


def copy_tsv(cohort, notes_dir, disclosure_dir):
    for item in ['note', 'order_narrative', 'order_impression', 'order_result_comment']:
        dir = cohort + '_' + item
        if (os.path.exists(os.path.join(notes_dir, dir))):
            file_list = [f for f in os.listdir(os.path.join(notes_dir, dir))]
            for f in file_list:
                shutil.copy(os.path.join(os.path.join(notes_dir, dir, f)), os.path.join(os.path.join(disclosure_dir, f)))
    return


def main():
    """
    """
    logging.info('START')
    logging.debug(f"""`base_dir` set to "{base_dir}".""")
    # pull metadata
    pull_metadata(note_version, id_type, 'note', sql_dir, data_dir, notes_dir, cohort)
    pull_metadata(note_version, id_type, 'order_narrative', sql_dir, data_dir, notes_dir, cohort)
    pull_metadata(note_version, id_type, 'order_impression', sql_dir, data_dir, notes_dir, cohort)
    pull_metadata(note_version, id_type, 'order_result_comment', sql_dir, data_dir, notes_dir, cohort)
    # split metadata into chunks, so that we can process data in chunks
    split_metadata('note', notes_dir, cohort)
    split_metadata('order_narrative', notes_dir, cohort)
    split_metadata('order_impression', notes_dir, cohort)
    split_metadata('order_result_comment', notes_dir, cohort)
    # pull text in parallel
    pull_text_in_parallel('note', sql_dir, notes_dir, cohort)
    pull_text_in_parallel('order_narrative', sql_dir, notes_dir, cohort)
    pull_text_in_parallel('order_impression', sql_dir, notes_dir, cohort)
    pull_text_in_parallel('order_result_comment', sql_dir, notes_dir, cohort)
    if (deid_mode == 'phi'):
        # copy note_metadata file to disclosure folder
        format_metadata_files('note', notes_dir, disclosure_dir)
        # combine all order metadata files and copy to discosure folder
        combine_order_metadata(notes_dir, cohort)
        format_metadata_files('order', notes_dir, disclosure_dir)
        # copy .tsv files with free text
        copy_tsv(cohort, notes_dir, disclosure_dir)
    else:
        # generate mappings
        generate_map(deid_mode, data_dir, map_dir, 'patient', cohort)
        generate_map(deid_mode, notes_dir, map_dir, 'encounter', cohort)
        generate_map(deid_mode, notes_dir, map_dir, 'note', cohort)
        generate_map(deid_mode, notes_dir, map_dir, 'note_link', cohort)
        generate_map(deid_mode, notes_dir, map_dir, 'order', cohort)
        generate_map(deid_mode, notes_dir, map_dir, 'provider', cohort)
        # deidentify metadata. We can create deidentified dataset or limited dataset.
        deid_metadata(deid_mode, 'note', map_dir, notes_dir, disclosure_dir)
        deid_metadata(deid_mode, 'order', map_dir, notes_dir, disclosure_dir)
        # prepare for text deidentification, i.e., deidentify ID in .tsv file(s)
        deid_tsv_note(map_dir, notes_dir)
        deid_tsv_order(map_dir, notes_dir)
    logging.info("END")



"""
QA: Make sure values were correctly converted.
"""

import argparse
import logging
import os
import pprint
from pathlib import Path
from typing_extensions import List, Union
# Third-party packages
import pandas as pd
# Local packages
from drapi.code.drapi.drapi import (choosePathToLog,
                                    getMapType,
                                    getTimestamp,
                                    makeDirPath,
                                    successiveParents)


# Functions
def iterateOverFiles(listOfFiles: List[Union[str, Path]],
                     SCRIPT_TEST_MODE: bool,
                     directoryName: Union[str, None] = None) -> pd.DataFrame:
    """
    """
    di = {}
    it1Total = len(listOfFiles)
    for it1, fpath in enumerate(listOfFiles, start=1):
        logger.info(f"""  Working on file {it1:,} of {it1Total:,}: "{fpath.name}".""")
        if SCRIPT_TEST_MODE:
            pass
        else:
            df = pd.read_csv(fpath, low_memory=False)
            result = getMapType(df=df)
            di[it1] = {"Directory": directoryName,
                       "File Name": fpath.name,
                       "Result": result}
    dfresult = pd.DataFrame.from_dict(data=di, orient="index")
    return dfresult


def qaListOfFilesAndDirectories(LIST_OF_FILES: List[Union[str, Path]],
                                LIST_OF_DIRECTORIES: List[Union[str, Path]],
                                SCRIPT_TEST_MODE: bool):
    """
    """
    # Sort input alphabetically for easier debugging.
    if LIST_OF_FILES:
        listOfFiles1 = sorted([Path(fpath) for fpath in LIST_OF_FILES])
    else:
        listOfFiles1 = []
    if LIST_OF_DIRECTORIES:
        listOfDirectories = sorted([Path(dpath) for dpath in LIST_OF_DIRECTORIES])
    else:
        listOfDirectories = []

    # Work on files first
    dfresult1 = iterateOverFiles(listOfFiles=listOfFiles1,
                                 SCRIPT_TEST_MODE=SCRIPT_TEST_MODE,
                                 directoryName=None)

    # Work on directories
    resultsli = []
    for dpath in listOfDirectories:
        logger.info(f"""  Working on directory "{dpath.name}".""")
        listOfFiles2 = sorted(list(dpath.iterdir()))
        dfresult2 = iterateOverFiles(listOfFiles=listOfFiles2,
                                     SCRIPT_TEST_MODE=SCRIPT_TEST_MODE,
                                     directoryName=dpath.name)
        resultsli.append(dfresult2)

    dfresult = pd.concat([dfresult1] + resultsli, axis=0)
    return dfresult


def combineMaps(path: Path,
                runIntermediateDir: Path,
                runOutputDir: Path,
                MESSAGE_MODULO_CHUNKS: int,
                rootPath: Path,
                logger: logging.Logger) -> List[Path]:
    """
    Assumes `PATH` is the "Maps by Portion" folder with the following structure
    - Maps by Portion
      - Portion 1
        - Variable 1
          - CSV File 1
          - CSV File 2
          - etc.
        - Variable 2
          - CSV File 1
          - CSV File 2
          - etc.
      - Portion 2
        - Variable 1
          - CSV File 1
          - CSV File 2
          - etc.
        - Variable 2
          - CSV File 1
          - CSV File 2
          - etc.
        - etc.
      - etc.
    """
    listOfPortions = sorted(list(path.iterdir()))
    dictOfVariables = {}
    it1 = 1
    for portionDir in listOfPortions:
        listOfVariables = sorted(list(portionDir.iterdir()))
        for variableDir in listOfVariables:
            variableDirName = variableDir.name
            listOfFiles = sorted(list(variableDir.iterdir()))
            diNew = {it: {"Variable Name": variableDirName,
                          "File Path": fpath} for it, fpath in enumerate(listOfFiles, start=it1)}
            it1 += len(listOfFiles)
            dictOfVariables.update(diNew)

    dfAllPaths = pd.DataFrame.from_dict(data=dictOfVariables, orient="index")
    dfAllPaths.index.name = "Index"
    dfAllPaths = dfAllPaths.sort_values(by=["Variable Name", "File Path"])
    savePathAllPaths = runOutputDir.joinpath("Paths for All Maps.CSV")
    dfAllPaths.to_csv(path_or_buf=savePathAllPaths)

    # Combine maps by variable name
    listOfVariableNames = dfAllPaths["Variable Name"].drop_duplicates()
    it2Total = len(listOfVariableNames)
    CHUNKSIZE = 50000
    savePathDir = runIntermediateDir.joinpath("Combined Maps", "By Variable")
    savePathDir.mkdir(mode=744, parents=True)
    listOfPathsForCombinedMaps = []
    for it2, variableName in enumerate(listOfVariableNames, start=1):
        logger.info(f"""  Working on variable {it2:,} of {it2Total}: "{variableName}".""")
        mask = dfAllPaths["Variable Name"] == variableName
        listOfPaths = dfAllPaths["File Path"][mask].to_list()
        it3Total = len(listOfPaths)
        listOfDataFrames = []
        for it3, fpath in enumerate(listOfPaths, start=1):
            logger.info(f"""    Working on file {it3:,} of {it3Total}: "{choosePathToLog(path=fpath, rootPath=rootPath)}".""")
            chunkGenerator1 = pd.read_csv(fpath, chunksize=CHUNKSIZE)
            chunkGenerator2 = pd.read_csv(fpath, chunksize=CHUNKSIZE)
            logger.info("    Reading the file to count the number of chunks.")
            it4Total = sum([1 for _ in chunkGenerator1])
            logger.info(f"    Reading the file to count the number of chunks - done. The file has {it4Total:,} chunks.")
            # Calculate logging frequency
            if it4Total < MESSAGE_MODULO_CHUNKS:
                moduloChunks = it4Total
            else:
                moduloChunks = round(it4Total / MESSAGE_MODULO_CHUNKS)
            if it4Total / moduloChunks < 100:
                moduloChunks = 1
            else:
                pass
            for it4, dfChunk in enumerate(chunkGenerator2, start=1):
                # Work on chunk
                if it4 % moduloChunks == 0:
                    allowLogging = True
                else:
                    allowLogging = False
                if allowLogging:
                    logger.info(f"""  ..  Working on chunk {it4:,} of {it4Total:,}.""")
                dfChunk = pd.DataFrame(dfChunk)  # For type hinting
                dfChunk = dfChunk.drop_duplicates()
                listOfDataFrames.append(dfChunk)
        dfAllUnique = pd.concat(listOfDataFrames)
        dfAllUnique = dfAllUnique.drop_duplicates()
        savePath = savePathDir.joinpath(f"{variableName}.CSV")
        logger.info(f"""    Saving combined maps to "{choosePathToLog(path=savePath, rootPath=rootDirectory)}".""")
        dfAllUnique.to_csv(path_or_buf=savePath,
                           index=False)
        listOfPathsForCombinedMaps.append(savePath)

    return listOfPathsForCombinedMaps


if __name__ == "__main__":
    # >>> `Argparse` arguments >>>
    parser = argparse.ArgumentParser()

    # Arguments: Main
    parser.add_argument("-d",
                        "--LIST_OF_DIRECTORIES",
                        nargs="*",
                        help="")
    parser.add_argument("-f",
                        "--LIST_OF_FILES",
                        nargs="*",
                        help="")

    parser.add_argument("-c",
                        "--COMBINE_MAPS",
                        help="This is the path to the directory that contains the portions and their respective variables. When this option is used all maps are combined by variable name and the analysis is performed on the resulting tables. Input is a string that is converted to a Path object.",
                        type=Path)

    parser.add_argument("-t",
                        "--SCRIPT_TEST_MODE",
                        help="",
                        action="store_true")

    # Arguments: General
    parser.add_argument("--CHUNKSIZE",
                        default=50000,
                        type=int,
                        help="The number of rows to read at a time from the CSV using Pandas `chunksize`")
    parser.add_argument("--MESSAGE_MODULO_CHUNKS",
                        default=50,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_CHUNKS`")
    parser.add_argument("--MESSAGE_MODULO_FILES",
                        default=100,
                        type=int,
                        help="How often to print a log message, i.e., print a message every x number of chunks, where x is `MESSAGE_MODULO_FILES`")

    # Arguments: Meta-variables
    parser.add_argument("--PROJECT_DIR_DEPTH",
                        default=2,
                        type=int,
                        help="")
    parser.add_argument("--DATA_REQUEST_DIR_DEPTH",
                        default=4,
                        type=int,
                        help="")
    parser.add_argument("--IRB_DIR_DEPTH",
                        default=3,
                        type=int,
                        help="")
    parser.add_argument("--IDR_DATA_REQUEST_DIR_DEPTH",
                        default=6,
                        type=int,
                        help="")
    parser.add_argument("--ROOT_DIRECTORY",
                        default="IRB_DIRECTORY",
                        type=str,
                        choices=["DATA_REQUEST_DIRECTORY",
                                 "IDR_DATA_REQUEST_DIRECTORY",
                                 "IRB_DIRECTORY",
                                 "PROJECT_OR_PORTION_DIRECTORY"],
                        help="")
    parser.add_argument("--LOG_LEVEL",
                        default=10,
                        type=int,
                        help="""Increase output verbosity. See "logging" module's log level for valid values.""")

    # Arguments: SQL connection settings
    parser.add_argument("--SERVER",
                        default="DWSRSRCH01.shands.ufl.edu",
                        type=str,
                        choices=["Acuo03.shands.ufl.edu",
                                 "EDW.shands.ufl.edu",
                                 "DWSRSRCH01.shands.ufl.edu",
                                 "IDR01.shands.ufl.edu",
                                 "RDW.shands.ufl.edu"],
                        help="")
    parser.add_argument("--DATABASE",
                        default="DWS_PROD",
                        type=str,
                        choices=["DWS_NOTES",
                                 "DWS_OMOP_PROD",
                                 "DWS_OMOP",
                                 "DWS_PROD"],  # TODO Add the i2b2 databases... or all the other databases?
                        help="")
    parser.add_argument("--USER_DOMAIN",
                        default="UFAD",
                        type=str,
                        choices=["UFAD"],
                        help="")
    parser.add_argument("--USERNAME",
                        default=os.environ["USER"],
                        type=str,
                        help="")
    parser.add_argument("--USER_ID",
                        default=None,
                        help="")
    parser.add_argument("--USER_PWD",
                        default=None,
                        help="")

    argNamespace = parser.parse_args()

    # Parsed arguments: Main
    LIST_OF_DIRECTORIES = argNamespace.LIST_OF_DIRECTORIES
    LIST_OF_FILES = argNamespace.LIST_OF_FILES

    COMBINE_MAPS = argNamespace.COMBINE_MAPS

    SCRIPT_TEST_MODE = argNamespace.SCRIPT_TEST_MODE

    # Parsed arguments: General
    CHUNKSIZE = argNamespace.CHUNKSIZE
    MESSAGE_MODULO_CHUNKS = argNamespace.MESSAGE_MODULO_CHUNKS
    MESSAGE_MODULO_FILES = argNamespace.MESSAGE_MODULO_FILES

    # Parsed arguments: Meta-variables
    PROJECT_DIR_DEPTH = argNamespace.PROJECT_DIR_DEPTH
    DATA_REQUEST_DIR_DEPTH = argNamespace.DATA_REQUEST_DIR_DEPTH
    IRB_DIR_DEPTH = argNamespace.IRB_DIR_DEPTH
    IDR_DATA_REQUEST_DIR_DEPTH = argNamespace.IDR_DATA_REQUEST_DIR_DEPTH

    ROOT_DIRECTORY = argNamespace.ROOT_DIRECTORY
    LOG_LEVEL = argNamespace.LOG_LEVEL

    # Parsed arguments: SQL connection settings
    SERVER = argNamespace.SERVER
    DATABASE = argNamespace.DATABASE
    USER_DOMAIN = argNamespace.USER_DOMAIN
    USERNAME = argNamespace.USERNAME
    USER_ID = argNamespace.USER_ID
    USER_PWD = argNamespace.USER_PWD
    # <<< `Argparse` arguments <<<

    # >>> Custom argument parsing >>>
    if LIST_OF_FILES or LIST_OF_DIRECTORIES:
        pass
    elif COMBINE_MAPS:
        pass
    else:
        parser.error("Although `LIST_OF_FILES`, `LIST_OF_DIRECTORIES`, and `COMBINE_MAPS` are marked as optional in the help text, you must actually provide arguments for at least one of them.")

    # <<< Custom argument parsing <<<

    # Variables: Path construction: General
    runTimestamp = getTimestamp()
    thisFilePath = Path(__file__)
    thisFileStem = thisFilePath.stem
    projectDir, _ = successiveParents(thisFilePath.absolute(), PROJECT_DIR_DEPTH)
    dataRequestDir, _ = successiveParents(thisFilePath.absolute(), DATA_REQUEST_DIR_DEPTH)
    IRBDir, _ = successiveParents(thisFilePath.absolute(), IRB_DIR_DEPTH)
    IDRDataRequestDir, _ = successiveParents(thisFilePath.absolute(), IDR_DATA_REQUEST_DIR_DEPTH)
    dataDir = projectDir.joinpath("data")
    if dataDir:
        inputDataDir = dataDir.joinpath("input")
        outputDataDir = dataDir.joinpath("output")
        if outputDataDir:
            runIntermediateDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
            runOutputDir = outputDataDir.joinpath(thisFileStem, runTimestamp)
    logsDir = projectDir.joinpath("logs")
    if logsDir:
        runLogsDir = logsDir.joinpath(thisFileStem)
    sqlDir = projectDir.joinpath("sql")

    if ROOT_DIRECTORY == "PROJECT_OR_PORTION_DIRECTORY":
        rootDirectory = projectDir
    elif ROOT_DIRECTORY == "DATA_REQUEST_DIRECTORY":
        rootDirectory = dataRequestDir
    elif ROOT_DIRECTORY == "IRB_DIRECTORY":
        rootDirectory = IRBDir
    elif ROOT_DIRECTORY == "IDR_DATA_REQUEST_DIRECTORY":
        rootDirectory = IDRDataRequestDir
    else:
        raise Exception("An unexpected error occurred.")

    # Variables: Path construction: Project-specific
    pass

    # Variables: SQL Parameters
    if USER_ID:
        uid = USER_ID[:]
    else:
        uid = fr"{USER_DOMAIN}\{USERNAME}"
    if USER_PWD:
        userPwd = USER_PWD
    else:
        userPWD = os.environ["HFA_UFADPWD"]
    conStr = f"mssql+pymssql://{uid}:{userPWD}@{SERVER}/{DATABASE}"

    # Variables: Other
    pass

    # Directory creation: General
    makeDirPath(runIntermediateDir)
    makeDirPath(runOutputDir)
    makeDirPath(runLogsDir)

    # Logging block
    logpath = runLogsDir.joinpath(f"log {runTimestamp}.log")
    logFormat = logging.Formatter("""[%(asctime)s][%(levelname)s](%(funcName)s): %(message)s""")

    logger = logging.getLogger(__name__)

    fileHandler = logging.FileHandler(logpath)
    fileHandler.setLevel(9)
    fileHandler.setFormatter(logFormat)

    streamHandler = logging.StreamHandler()
    streamHandler.setLevel(LOG_LEVEL)
    streamHandler.setFormatter(logFormat)

    logger.addHandler(fileHandler)
    logger.addHandler(streamHandler)

    logger.setLevel(9)

    logger.info(f"""Begin running "{thisFilePath}".""")
    logger.info(f"""All other paths will be reported in debugging relative to `{ROOT_DIRECTORY}`: "{rootDirectory}".""")
    logger.info(f"""Script arguments:


    # Arguments
    `CHUNKSIZE`: "{CHUNKSIZE}"
    `MESSAGE_MODULO_CHUNKS`: "{MESSAGE_MODULO_CHUNKS}"
    `MESSAGE_MODULO_FILES`: "{MESSAGE_MODULO_FILES}"

    # Arguments: General
    `PROJECT_DIR_DEPTH`: "{PROJECT_DIR_DEPTH}" ----------> "{projectDir}"
    `IRB_DIR_DEPTH`: "{IRB_DIR_DEPTH}" --------------> "{IRBDir}"
    `IDR_DATA_REQUEST_DIR_DEPTH`: "{IDR_DATA_REQUEST_DIR_DEPTH}" -> "{IDRDataRequestDir}"

    `LOG_LEVEL` = "{LOG_LEVEL}"

    # Arguments: SQL connection settings
    `SERVER` = "{SERVER}"
    `DATABASE` = "{DATABASE}"
    `USER_DOMAIN` = "{USER_DOMAIN}"
    `USERNAME` = "{USERNAME}"
    `USER_ID` = "{USER_ID}"
    `USER_PWD` = censored
    """)
    argList = argNamespace._get_args() + argNamespace._get_kwargs()
    argListString = pprint.pformat(argList)
    logger.info(f"""Script arguments:\n{argListString}""")

    # >>> Begin module body >>>
    if LIST_OF_FILES or LIST_OF_DIRECTORIES:
        dfresult = qaListOfFilesAndDirectories(LIST_OF_DIRECTORIES=LIST_OF_DIRECTORIES,
                                               LIST_OF_FILES=LIST_OF_FILES,
                                               SCRIPT_TEST_MODE=SCRIPT_TEST_MODE)
    elif COMBINE_MAPS:
        listOfPathsForCombinedMaps = combineMaps(path=COMBINE_MAPS,
                                                 runIntermediateDir=runIntermediateDir,
                                                 runOutputDir=runOutputDir,
                                                 MESSAGE_MODULO_CHUNKS=MESSAGE_MODULO_CHUNKS,
                                                 rootPath=rootDirectory,
                                                 logger=logger)
        dfresult = qaListOfFilesAndDirectories(LIST_OF_DIRECTORIES=[],
                                               LIST_OF_FILES=listOfPathsForCombinedMaps,
                                               SCRIPT_TEST_MODE=SCRIPT_TEST_MODE)

    # Report results
    if SCRIPT_TEST_MODE:
        pass
    else:
        dfresult = dfresult.sort_values(by=["Directory",
                                            "File Name"])
        savePath = runOutputDir.joinpath("Map Type Results.CSV")
        dfresult.to_csv(savePath)
        dfString = dfresult.to_string()

        logger.info(f"""Map Type Results:\n{dfString}""")

    # <<< End module body <<<
    logger.info(f"""Finished running "{thisFilePath.absolute().relative_to(rootDirectory)}".""")
